from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.service.seqdb import BaseSeqdbService
from gen_epix.fastapp import App
from gen_epix.seqdb.domain.command import (
    RetrievePhylogeneticTreeCommand as SeqdbRetrievePhylogeneticTreeCommand,
)
from gen_epix.seqdb.domain.enum import Role as SeqdbRole
from gen_epix.seqdb.domain.enum import TreeAlgorithm as SeqdbTreeAlgorithm
from gen_epix.seqdb.domain.model import PhylogeneticTree as SeqdbPhylogeneticTree
from gen_epix.seqdb.domain.model import User as SeqdbUser


class SeqdbService(BaseSeqdbService):

    def __init__(self, app: App, ext_app: App, **kwargs: dict) -> None:
        super().__init__(app, **kwargs)
        self._ext_app = ext_app
        # TODO: get user from config data
        self._functional_user = SeqdbUser(
            id=UUID("00000000-0000-0000-0000-000000000001"),
            email="functional.user@org.org",
            roles={SeqdbRole.ADMIN},
            data_collection_ids=set(),
        )

    @property
    def ext_app(self) -> App:
        return self._ext_app

    @property
    def functional_user(self) -> SeqdbUser:
        return self._functional_user

    def retrieve_phylogenetic_tree(
        self, cmd: command.RetrievePhylogeneticTreeBySequencesCommand
    ) -> model.PhylogeneticTree | None:
        user = cmd.user
        leaf_id_mapper = cmd.props.get("leaf_id_mapper")
        if leaf_id_mapper:
            leaf_names = [str(leaf_id_mapper(x)) for x in cmd.sequence_ids]
        else:
            leaf_names = None
        seqdb_cmd = SeqdbRetrievePhylogeneticTreeCommand(
            user=self.functional_user,
            seq_distance_protocol_id=cmd.seqdb_seq_distance_protocol_id,
            tree_algorithm=SeqdbTreeAlgorithm[cmd.tree_algorithm_code.value],
            seq_ids=cmd.sequence_ids,
            leaf_names=leaf_names,
        )
        seqdb_phylogenetic_tree: SeqdbPhylogeneticTree = self.ext_app.handle(seqdb_cmd)
        phylogenetic_tree = model.PhylogeneticTree(
            tree_algorithm_code=cmd.tree_algorithm_code,
            sequence_ids=seqdb_phylogenetic_tree.seq_ids,
            leaf_ids=(
                [UUID(x) for x in seqdb_phylogenetic_tree.leaf_names]
                if seqdb_phylogenetic_tree.leaf_names
                else None
            ),
            newick_repr=seqdb_phylogenetic_tree.newick_repr,
        )
        return phylogenetic_tree

    def retrieve_genetic_sequences(self, cmd) -> list[model.GeneticSequence]:
        raise NotImplementedError()

    # def retrieve_allele_profile(
    #     self,
    #     cmd: command.RetrieveAlleleProfileCommand,
    # ) -> model.SeqDbAlleleProfile:
    #     raise NotImplementedError()



import json
import logging
import uuid
from datetime import datetime
from typing import Type

import httpx
from fastapi import Request
from fastapi.openapi.models import OAuth2, OAuthFlowAuthorizationCode, OAuthFlows
from fastapi.security.open_id_connect_url import OpenIdConnect
from fastapi.security.utils import get_authorization_scheme_param
from jose import ExpiredSignatureError, JWTError, jwk, jwt
from jose.backends.base import Key
from jose.exceptions import JWTClaimsError

from gen_epix.fastapp import exc
from gen_epix.fastapp.enum import AuthProtocol, OauthFlowType
from gen_epix.fastapp.log import BaseLogItem, LogItem
from gen_epix.fastapp.services.auth.idp_client import IDPClient
from gen_epix.fastapp.services.auth.model import (
    Claims,
    IdentityProvider,
    OIDCConfiguration,
)


class OIDCClient(IDPClient, OpenIdConnect):

    def __init__(
        self,
        oidc_configuration: OIDCConfiguration,
        logger: logging.Logger | None = None,
        log_item_class: Type[BaseLogItem] = LogItem,
        **kwargs: dict,
    ):
        self._id = kwargs.get("id", uuid.uuid4())
        # Set input properties and initialize some
        self._logger = logger
        self._log_item_class = log_item_class
        self._cfg = oidc_configuration
        self._signing_keys: dict[str, Key] = {}

        # Retrieve remaining information
        self._init_from_discovery_url(
            discovery_document=kwargs.get("discovery_document")
        )
        self._signing_keys = {}
        # self._load_keys()
        flows = OAuthFlows()
        flows.authorizationCode = OAuthFlowAuthorizationCode(
            authorizationUrl=self._authorization_endpoint,
            tokenUrl=self._token_endpoint,
            scopes={k: k for k in self._cfg.scope.split()},
        )

        # Set SecurityBase properties
        self.model = OAuth2(flows=flows)
        self.scheme_name = self._issuer
        self.token_name = "id_token"

    @property
    def issuer(self) -> str:
        return self._issuer

    @property
    def audience(self) -> str:
        return self._cfg.client_id

    @property
    def id(self) -> uuid.UUID:
        return self._id

    def _should_verify_ssl(self, url: str) -> bool:
        return "127.0.0.1" not in url and "localhost" not in url

    def _init_from_discovery_url(self, discovery_document: dict | None = None) -> None:
        try:
            if discovery_document is None:
                with httpx.Client(
                    verify=self._should_verify_ssl(self._cfg.discovery_url)
                ) as client:
                    response = client.get(self._cfg.discovery_url)
                    discovery_document = response.json()
            self._issuer = discovery_document["issuer"]
            self._authorization_endpoint = discovery_document["authorization_endpoint"]
            self._token_endpoint = discovery_document["token_endpoint"]
            self._jwks_uri = discovery_document["jwks_uri"]
            self._userinfo_endpoint = discovery_document["userinfo_endpoint"]
            self._response_types_supported = discovery_document[
                "response_types_supported"
            ]
            self._subject_types_supported = discovery_document[
                "subject_types_supported"
            ]
            self._id_token_signing_alg_values_supported = discovery_document[
                "id_token_signing_alg_values_supported"
            ]
        except Exception as exception:
            msg = f"Error access discovery URL for OIDC service: {exception}"
            if self._logger:
                self._logger.error(
                    self._log_item_class(
                        code="cfe970aa", msg=msg, exception=exception
                    ).dumps()
                )
            raise exc.InitializationServiceError(msg) from exception

    async def get_jwk_from_jwt_token(self, jwt_token: str) -> Key:
        try:
            header = jwt.get_unverified_header(jwt_token)
        except JWTError as e:
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="4cff1367",
                        msg="Unable to parse header from token",
                        jwt=jwt_token,
                        exception=e,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError() from e

        key_id = header.get("kid")
        if not key_id:
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="0184bc35",
                        msg="No key ID found in token header",
                        jwt=jwt_token,
                    ).dumps()
                )
            raise exc.UnauthorizedAuthError()

        # Verify that the signing key in this session is outdated, fetch new one if so
        # TODO: verify if fetching new signing keys is ok
        key = self._signing_keys.get(key_id)
        if not key:
            if self._logger:
                self._logger.info(
                    self._log_item_class(
                        code="e90dd1aa",
                        msg="Key ID not found among signing keys, fetching new ones",
                        jwt=jwt_token,
                        key_id=key_id,
                    ).dumps()
                )
            self._load_keys()
            key = self._signing_keys.get(key_id)
            if not key:
                if self._logger:
                    self._logger.warning(
                        self._log_item_class(
                            code="2a5975ff",
                            msg="Key ID not found amoung newly fetched signing keys",
                            key_id=key_id,
                        ).dumps()
                    )
                raise exc.UnauthorizedAuthError()
            if self._logger:
                self._logger.info(
                    self._log_item_class(
                        code="c448ead5",
                        msg="Key ID found among newly fetched signing keys",
                        jwt=jwt_token,
                        key_id=key_id,
                    ).dumps()
                )
        return key

    async def get_claims_from_token(
        self, jwt_token: str
    ) -> dict[str, str | int | bool | list[str]] | None:
        # Decode token without verifying signature to make sure this token is generated
        # by this OIDC server
        claims = jwt.get_unverified_claims(jwt_token)
        if claims["iss"] != self._issuer or claims.get("aud") != self._cfg.client_id:
            # Different OIDC server
            return None

        iat: int = claims.get("iat", -1)
        if iat == -1 or iat > int(datetime.now().timestamp()):
            # Token issued in the future
            return None

        # Get key to verify signature and decode again
        key = await self.get_jwk_from_jwt_token(jwt_token)
        try:
            claims = jwt.decode(
                jwt_token,
                key=key,
                algorithms=self._id_token_signing_alg_values_supported,
                audience=self._cfg.client_id,
                issuer=self._issuer,
                # TODO: Check if this is not a security risk
                options={"verify_at_hash": False},
            )
        except Exception as exception:
            msg = "Unable to decode JWT: "
            if isinstance(exception, ExpiredSignatureError):
                msg += "signature has expired"
            elif isinstance(exception, JWTClaimsError):
                msg += "some claims are invalid"
            elif isinstance(exception, JWTError):
                msg += "signature is invalid"
            else:
                msg += "unknown issue"
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="f4b73564",
                        msg=msg,
                        jwt=jwt_token,
                        exception=exception,
                    ).dumps()
                )
            raise exc.CredentialsAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            ) from exception

        issuer = claims["iss"]
        sub = claims.get("sub")
        if not issuer or not sub:
            if not issuer and not sub:
                msg_part = "no issuer and no sub"
            elif issuer and not sub:
                msg_part = "no sub"
            else:
                msg_part = "no issuer"
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="b4a1d49b",
                        msg=f"JWT does not contain required claims: {msg_part}",
                        jwt=jwt_token,
                    ).dumps()
                )
            raise exc.CredentialsAuthError(
                http_props={"headers": {"WWW-Authenticate": "Bearer"}}
            )
        return claims

    def get_claims_from_userinfo(
        self, access_token: str
    ) -> dict[str, str | int | bool | list[str]]:
        try:
            with httpx.Client(
                verify=self._should_verify_ssl(self._userinfo_endpoint)
            ) as client:
                response = client.get(
                    self._userinfo_endpoint,
                    headers={"Authorization": f"Bearer {access_token}"},
                )
                claims = json.loads(response.content)
                if not isinstance(claims, dict) or "error" in claims:
                    # Currently e.g. "InvalidAuthenticationToken"
                    if self._logger:
                        self._logger.warning(
                            self._log_item_class(
                                code="ce05d050",
                                msg=f"Unable to get claims from {self._userinfo_endpoint}: claims contain error",
                                claims=claims,
                            ).dumps()
                        )
                    raise exc.ServiceUnavailableError()
                return claims
        except Exception as exception:
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="ac6c84f7",
                        msg=f"Unable to get claims from {self._userinfo_endpoint}",
                        exception=exception,
                    ).dumps()
                )
            return {}

    def get_identity_provider(self) -> IdentityProvider:
        return IdentityProvider(
            name=self._cfg.name,
            label=self._cfg.label,
            client_id=self._cfg.client_id,
            client_secret=self._cfg.client_secret,
            discovery_url=self._cfg.discovery_url,
            issuer=self._issuer,
            auth_protocol=AuthProtocol.OIDC,
            oauth_flow=OauthFlowType.AUTHORIZATION_CODE,
            scope=self._cfg.scope,
        )

    def _load_keys(self) -> None:
        try:
            with httpx.Client(verify=self._should_verify_ssl(self._jwks_uri)) as client:
                # get keys
                response = client.get(self._jwks_uri)
                response.raise_for_status()
                response_dict = response.json()

                self._signing_keys = {
                    key_data["kid"]: jwk.construct(
                        key_data=key_data,
                        algorithm=key_data.get(
                            "alg", "RS256"
                        ),  # Assume RS256 if alg is not specified
                    )
                    for key_data in response_dict["keys"]
                    if key_data["use"] == "sig"
                }
        except Exception as exception:
            if self._logger:
                self._logger.warning(
                    self._log_item_class(
                        code="edab2e97",
                        msg=f"Unable to load new signing keys from {self._jwks_uri}",
                        exception=exception,
                    ).dumps()
                )
            raise exc.ServiceUnavailableError() from exception

    async def __call__(self, request: Request) -> Claims | None:  # type: ignore
        """
        Retrieve verified claims for the user based on the request.
        """
        if authorization := request.headers.get("authorization"):
            scheme, token = get_authorization_scheme_param(authorization)
            if scheme.upper() == "BEARER":
                # TODO: check if this is a security risk
                # or whether it should return an error
                try:
                    claims = await self.get_claims_from_token(token)
                    if not claims:
                        return None
                    return Claims(
                        claims=claims, scheme=scheme, token=token, idp_client_id=self.id
                    )
                except exc.AuthException as exception:
                    if self._logger:
                        self._logger.warning(
                            self._log_item_class(
                                code="ac521d94",
                                exception=exception,
                            ).dumps()
                        )
                    return None

            else:
                # Authorization scheme not implemented
                if self._logger:
                    self._logger.warning(
                        self._log_item_class(
                            code="ecb88df4",
                            msg=f"Authorization scheme {scheme} not implemented",
                        ).dumps()
                    )
                return None
        if self._logger:
            self._logger.warning(
                self._log_item_class(
                    code="e1dad160",
                    msg="No authorisation information provided in header",
                ).dumps()
            )
        return None



# This file is automatically generated by generate_pydantic_model_defs.py.
# It needs to be copied to gen_epix/omopdb/domain/model/omop.py.
# The order of certain code elements may need to be changed manually.

from __future__ import annotations

from datetime import date, datetime
from typing import Any, ClassVar
from uuid import UUID

from pydantic import Field, RootModel

from gen_epix.fastapp.domain import Entity, create_keys
from gen_epix.omopdb.domain import DOMAIN, enum
from gen_epix.omopdb.domain.model.base import Model

_SERVICE_TYPE = enum.ServiceType.OMOP
_ENTITY_KWARGS = {
    "service_type": _SERVICE_TYPE,
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class ModelModel(RootModel[Any]):
    root: Any


class Person(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Persons",
        table_name="person",
        persistable=True,
        keys=create_keys({1: "person_id"}),
        **_ENTITY_KWARGS,
    )

    care_site_id: UUID | None = Field(
        None, description="care_site_id", title="Care Site Id"
    )
    day_of_birth: int | None = Field(
        None, description="day_of_birth", ge=1, le=31, title="Day Of Birth"
    )
    ethnicity_concept_id: int | None = Field(
        None,
        description="ethnicity_concept_id",
        le=2147483647,
        title="Ethnicity Concept Id",
    )
    ethnicity_source_concept_id: int | None = Field(
        None,
        description="ethnicity_source_concept_id",
        le=2147483647,
        title="Ethnicity Source Concept Id",
    )
    ethnicity_source_value: str | None = Field(
        None,
        description="ethnicity_source_value",
        max_length=50,
        title="Ethnicity Source Value",
    )
    birth_datetime: datetime | None = Field(
        None, description="birth_datetime", title="Birth Datetime"
    )
    death_datetime: datetime | None = Field(
        None, description="death_datetime", title="Death Datetime"
    )
    gender_concept_id: int | None = Field(
        None, description="gender_concept_id", le=2147483647, title="Gender Concept Id"
    )
    gender_source_concept_id: int | None = Field(
        None,
        description="gender_source_concept_id",
        le=2147483647,
        title="Gender Source Concept Id",
    )
    gender_source_value: str | None = Field(
        None,
        description="gender_source_value",
        max_length=50,
        title="Gender Source Value",
    )
    location_id: UUID | None = Field(
        None, description="location_id", title="Location Id"
    )
    month_of_birth: int | None = Field(
        None, description="month_of_birth", ge=1, le=12, title="Month Of Birth"
    )
    person_id: UUID = Field(
        ..., description="person_id, PRIMARY KEY", title="Person Id"
    )
    person_source_value: str | None = Field(
        None,
        description="person_source_value",
        max_length=50,
        title="Person Source Value",
    )
    person_type_concept_id: int = Field(
        ...,
        description="person_type_concept_id",
        le=2147483647,
        title="Person Type Concept Id",
    )
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    race_concept_id: int | None = Field(
        None, description="race_concept_id", le=2147483647, title="Race Concept Id"
    )
    race_source_concept_id: int | None = Field(
        None,
        description="race_source_concept_id",
        le=2147483647,
        title="Race Source Concept Id",
    )
    race_source_value: str | None = Field(
        None, description="race_source_value", max_length=50, title="Race Source Value"
    )
    year_of_birth: int | None = Field(
        None, description="year_of_birth", title="Year Of Birth"
    )
    provenance_id: UUID = Field(
        ..., description="provenance_id", title="Provenance Id"
    )  # should be for every omop_subject unique
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class ObservationPeriod(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ObservationPeriods",
        table_name="observation_period",
        persistable=True,
        keys=create_keys({1: "observation_period_id"}),
        **_ENTITY_KWARGS,
    )

    period_type_concept_id: int = Field(
        ...,
        description="period_type_concept_id",
        le=2147483647,
        title="Period Type Concept Id",
    )
    observation_period_end_date: date | None = Field(
        None,
        description="observation_period_end_date",
        title="Observation Period End Date",
    )
    observation_period_end_iso_interval: str = Field(
        ...,
        description="observation_period_end_iso_interval",
        max_length=55,
        title="Observation Period End Iso Interval",
    )
    observation_period_id: UUID = Field(
        ...,
        description="observation_period_id, PRIMARY KEY",
        title="Observation Period Id",
    )
    observation_period_start_date: date | None = Field(
        None,
        description="observation_period_start_date",
        title="Observation Period Start Date",
    )
    observation_period_start_iso_interval: str = Field(
        ...,
        description="observation_period_start_iso_interval",
        max_length=55,
        title="Observation Period Start Iso Interval",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")


class VisitOccurrence(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="VisitOccurrences",
        table_name="visit_occurrence",
        persistable=True,
        keys=create_keys({1: "visit_occurrence_id"}),
        **_ENTITY_KWARGS,
    )

    admitted_from_concept_id: int = Field(
        ...,
        description="admitted_from_concept_id",
        le=2147483647,
        title="Admitted From Concept Id",
    )
    discharge_to_concept_id: int = Field(
        ...,
        description="discharge_to_concept_id",
        le=2147483647,
        title="Discharge To Concept Id",
    )
    visit_concept_id: int = Field(
        ..., description="visit_concept_id", le=2147483647, title="Visit Concept Id"
    )
    visit_source_concept_id: int = Field(
        ...,
        description="visit_source_concept_id",
        le=2147483647,
        title="Visit Source Concept Id",
    )
    visit_type_concept_id: int = Field(
        ...,
        description="visit_type_concept_id",
        le=2147483647,
        title="Visit Type Concept Id",
    )
    care_site_id: UUID | None = Field(
        None, description="care_site_id", title="Care Site Id"
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    preceding_visit_occurrence_id: UUID | None = Field(
        None,
        description="preceding_visit_occurrence_id",
        title="Preceding Visit Occurrence Id",
    )
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_occurrence_id: UUID = Field(
        ..., description="visit_occurrence_id, PRIMARY KEY", title="Visit Occurrence Id"
    )
    visit_end_date: date | None = Field(
        None, description="visit_end_date", title="Visit End Date"
    )
    visit_end_datetime: datetime = Field(
        ..., description="visit_end_datetime", title="Visit End Datetime"
    )
    visit_start_date: date | None = Field(
        None, description="visit_start_date", title="Visit Start Date"
    )
    visit_start_datetime: datetime = Field(
        ..., description="visit_start_datetime", title="Visit Start Datetime"
    )
    visit_source_value: str | None = Field(
        None,
        description="visit_source_value",
        max_length=50,
        title="Visit Source Value",
    )
    admitted_from_source_value: str | None = Field(
        None,
        description="admitted_from_source_value",
        max_length=50,
        title="Admitted From Source Value",
    )
    discharge_to_source_value: str | None = Field(
        None,
        description="discharge_to_source_value",
        max_length=50,
        title="Discharge To Source Value",
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class VisitDetail(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="VisitDetails",
        table_name="visit_detail",
        persistable=True,
        keys=create_keys({1: "visit_detail_id"}),
        **_ENTITY_KWARGS,
    )

    admitted_from_source_value: int = Field(
        ...,
        description="admitted_from_source_value",
        le=2147483647,
        title="Admitted From Source Value",
    )
    discharge_to_concept_id: int = Field(
        ...,
        description="discharge_to_concept_id",
        le=2147483647,
        title="Discharge To Concept Id",
    )
    visit_detail_concept_id: int = Field(
        ...,
        description="visit_detail_concept_id",
        le=2147483647,
        title="Visit Detail Concept Id",
    )
    visit_detail_source_concept_id: int = Field(
        ...,
        description="visit_detail_source_concept_id",
        le=2147483647,
        title="Visit Detail Source Concept Id",
    )
    visit_detail_type_concept_id: int = Field(
        ...,
        description="visit_detail_type_concept_id",
        le=2147483647,
        title="Visit Detail Type Concept Id",
    )
    care_site_id: UUID | None = Field(
        None, description="care_site_id", title="Care Site Id"
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    preceding_visit_detail_id: UUID | None = Field(
        None, description="preceding_visit_detail_id", title="Preceding Visit Detail Id"
    )
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID = Field(
        ..., description="visit_detail_id, PRIMARY KEY", title="Visit Detail Id"
    )
    visit_detail_end_date: date = Field(
        ..., description="visit_detail_end_date", title="Visit Detail End Date"
    )
    visit_detail_end_datetime: datetime | None = Field(
        None, description="visit_detail_end_datetime", title="Visit Detail End Datetime"
    )
    visit_detail_start_date: date = Field(
        ..., description="visit_detail_start_date", title="Visit Detail Start Date"
    )
    visit_detail_start_datetime: datetime | None = Field(
        None,
        description="visit_detail_start_datetime",
        title="Visit Detail Start Datetime",
    )
    visit_detail_parent_id: UUID | None = Field(
        None, description="visit_detail_parent_id", title="Visit Detail Parent Id"
    )
    visit_detail_source_value: str | None = Field(
        None,
        description="visit_detail_source_value",
        max_length=50,
        title="Visit Detail Source Value",
    )
    admitted_from_concept_id: str | None = Field(
        None,
        description="admitted_from_concept_id",
        max_length=50,
        title="Admitted From Concept Id",
    )
    discharge_to_source_value: str | None = Field(
        None,
        description="discharge_to_source_value",
        max_length=50,
        title="Discharge To Source Value",
    )
    visit_occurrence_id: UUID = Field(
        ..., description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class ConditionOccurrence(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConditionOccurrences",
        table_name="condition_occurrence",
        persistable=True,
        keys=create_keys({1: "condition_occurrence_id"}),
        **_ENTITY_KWARGS,
    )

    condition_concept_id: int = Field(
        ...,
        description="condition_concept_id",
        le=2147483647,
        title="Condition Concept Id",
    )
    condition_end_date: date | None = Field(
        None, description="condition_end_date", title="Condition End Date"
    )
    condition_source_concept_id: int = Field(
        ...,
        description="condition_source_concept_id",
        le=2147483647,
        title="Condition Source Concept Id",
    )
    condition_occurrence_id: UUID = Field(
        ...,
        description="condition_occurrence_id, PRIMARY KEY",
        title="Condition Occurrence Id",
    )
    condition_status_concept_id: int = Field(
        ...,
        description="condition_status_concept_id",
        le=2147483647,
        title="Condition Status Concept Id",
    )
    condition_type_concept_id: int = Field(
        ...,
        description="condition_type_concept_id",
        le=2147483647,
        title="Condition Type Concept Id",
    )
    condition_end_datetime: datetime | None = Field(
        None, description="condition_end_datetime", title="Condition End Datetime"
    )
    condition_end_iso_interval: str | None = Field(
        None,
        description="condition_end_iso_interval",
        max_length=55,
        title="Condition End Iso Interval",
    )
    condition_start_date: date = Field(
        ..., description="condition_start_date", title="Condition Start Date"
    )
    condition_start_datetime: datetime = Field(
        ..., description="condition_start_datetime", title="Condition Start Datetime"
    )
    condition_start_iso_interval: str = Field(
        ...,
        description="condition_start_iso_interval",
        max_length=55,
        title="Condition Start Iso Interval",
    )
    stop_reason: str | None = Field(
        None, description="stop_reason", max_length=20, title="Stop Reason"
    )
    condition_source_value: str | None = Field(
        None,
        description="condition_source_value",
        max_length=50,
        title="Condition Source Value",
    )
    condition_status_source_value: str | None = Field(
        None,
        description="condition_status_source_value",
        max_length=50,
        title="Condition Status Source Value",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class DrugExposure(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="DrugExposures",
        table_name="drug_exposure",
        persistable=True,
        keys=create_keys({1: "drug_exposure_id"}),
        **_ENTITY_KWARGS,
    )

    drug_concept_id: int = Field(
        ..., description="drug_concept_id", le=2147483647, title="Drug Concept Id"
    )
    drug_source_concept_id: int = Field(
        ...,
        description="drug_source_concept_id",
        le=2147483647,
        title="Drug Source Concept Id",
    )
    drug_type_concept_id: int = Field(
        ...,
        description="drug_type_concept_id",
        le=2147483647,
        title="Drug Type Concept Id",
    )
    drug_exposure_end_date: date | None = Field(
        None, description="drug_exposure_end_date", title="Drug Exposure End Date"
    )
    drug_exposure_id: UUID = Field(
        ..., description="drug_exposure_id, PRIMARY KEY", title="Drug Exposure Id"
    )
    drug_exposure_end_datetime: datetime | None = Field(
        None,
        description="drug_exposure_end_datetime",
        title="Drug Exposure End Datetime",
    )
    drug_exposure_end_iso_interval: str | None = Field(
        None,
        description="drug_exposure_end_iso_interval",
        max_length=55,
        title="Drug Exposure End Iso Interval",
    )
    drug_exposure_start_date: date | None = Field(
        None, description="drug_exposure_start_date", title="Drug Exposure Start Date"
    )
    drug_exposure_start_datetime: datetime | None = Field(
        None,
        description="drug_exposure_start_datetime",
        title="Drug Exposure Start Datetime",
    )
    drug_exposure_start_iso_interval: str = Field(
        ...,
        description="drug_exposure_start_iso_interval",
        max_length=55,
        title="Drug Exposure Start Iso Interval",
    )
    verbatim_end_date: date | None = Field(
        None, description="verbatim_end_date", title="Verbatim End Date"
    )
    quantity: float | None = Field(None, description="quantity", title="Quantity")
    refills: int | None = Field(None, description="refills", title="Refills")
    days_supply: int | None = Field(
        None, description="days_supply", title="Days Supply"
    )
    stop_reason: str | None = Field(
        None, description="stop_reason", max_length=20, title="Stop Reason"
    )
    lot_number: str | None = Field(
        None, description="lot_number", max_length=50, title="Lot Number"
    )
    drug_source_value: str | None = Field(
        None, description="drug_source_value", max_length=50, title="Drug Source Value"
    )
    route_source_value: str | None = Field(
        None,
        description="route_source_value",
        max_length=50,
        title="Route Source Value",
    )
    dose_unit_source_value: str | None = Field(
        None,
        description="dose_unit_source_value",
        max_length=50,
        title="Dose Unit Source Value",
    )
    sig: str | None = Field(None, description="sig", title="Sig")
    route_concept_id: int | None = Field(
        None, description="route_concept_id", le=2147483647, title="Route Concept Id"
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class ProcedureOccurrence(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ProcedureOccurrences",
        table_name="procedure_occurrence",
        persistable=True,
        keys=create_keys({1: "procedure_occurrence_id"}),
        **_ENTITY_KWARGS,
    )

    modifier_concept_id: int | None = Field(
        None,
        description="modifier_concept_id",
        le=2147483647,
        title="Modifier Concept Id",
    )
    procedure_concept_id: int = Field(
        ...,
        description="procedure_concept_id",
        le=2147483647,
        title="Procedure Concept Id",
    )
    procedure_type_concept_id: int = Field(
        ...,
        description="procedure_type_concept_id",
        le=2147483647,
        title="Procedure Type Concept Id",
    )
    procedure_occurrence_id: UUID = Field(
        ...,
        description="procedure_occurrence_id, PRIMARY KEY",
        title="Procedure Occurrence Id",
    )
    provider_id: int | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_occurrence_id: int | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    visit_detail_id: int | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    procedure_date: date | None = Field(
        None, description="procedure_date", title="Procedure Date"
    )
    procedure_datetime: datetime | None = Field(
        None, description="procedure_datetime", title="Procedure Datetime"
    )
    quantity: int | None = Field(None, description="quantity", title="Quantity")
    procedure_source_concept_id: int = Field(
        ...,
        description="procedure_source_concept_id",
        le=2147483647,
        title="Procedure Source Concept Id",
    )
    procedure_source_value: str | None = Field(
        None,
        description="procedure_source_value",
        max_length=50,
        title="Procedure Source Value",
    )
    modifier_source_value: str | None = Field(
        None,
        description="modifier_source_value",
        max_length=50,
        title="Modifier Source Value",
    )
    procedure_iso_interval: str = Field(
        ...,
        description="procedure_iso_interval",
        max_length=55,
        title="Procedure Iso Interval",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class DeviceExposure(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="DeviceExposures",
        table_name="device_exposure",
        persistable=True,
        keys=create_keys({1: "device_exposure_id"}),
        **_ENTITY_KWARGS,
    )

    device_concept_id: int = Field(
        ..., description="device_concept_id", le=2147483647, title="Device Concept Id"
    )
    device_source_concept_id: int = Field(
        ...,
        description="device_source_concept_id",
        le=2147483647,
        title="Device Source Concept Id",
    )
    device_type_concept_id: int = Field(
        ...,
        description="device_type_concept_id",
        le=2147483647,
        title="Device Type Concept Id",
    )
    device_exposure_end_date: date | None = Field(
        None, description="device_exposure_end_date", title="Device Exposure End Date"
    )
    device_exposure_id: UUID = Field(
        ..., description="device_exposure_id, PRIMARY KEY", title="Device Exposure Id"
    )
    device_exposure_end_datetime: datetime | None = Field(
        None,
        description="device_exposure_end_datetime",
        title="Device Exposure End Datetime",
    )
    device_exposure_end_iso_interval: str | None = Field(
        None,
        description="device_exposure_end_iso_interval",
        max_length=55,
        title="Device Exposure End Iso Interval",
    )
    device_exposure_start_date: date = Field(
        ...,
        description="device_exposure_start_date",
        title="Device Exposure Start Date",
    )
    device_exposure_start_datetime: datetime = Field(
        ...,
        description="device_exposure_start_datetime",
        title="Device Exposure Start Datetime",
    )
    device_exposure_start_iso_interval: str = Field(
        ...,
        description="device_exposure_start_iso_interval",
        max_length=55,
        title="Device Exposure Start Iso Interval",
    )
    quantity: int | None = Field(None, description="quantity", title="Quantity")
    unique_device_id: str | None = Field(
        None, description="unique_device_id", max_length=50, title="Unique Device Id"
    )
    device_source_value: str | None = Field(
        None,
        description="device_source_value",
        max_length=50,
        title="Device Source Value",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class Measurement(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Measurements",
        table_name="measurement",
        persistable=True,
        keys=create_keys({1: "measurement_id"}),
        **_ENTITY_KWARGS,
    )

    measurement_concept_id: int = Field(
        ...,
        description="measurement_concept_id",
        le=2147483647,
        title="Measurement Concept Id",
    )
    measurement_source_concept_id: int = Field(
        ...,
        description="measurement_source_concept_id",
        le=2147483647,
        title="Measurement Source Concept Id",
    )
    measurement_type_concept_id: int = Field(
        ...,
        description="measurement_type_concept_id",
        le=2147483647,
        title="Measurement Type Concept Id",
    )
    measurement_date: date = Field(
        ..., description="measurement_date", title="Measurement Date"
    )
    measurement_iso_interval: str = Field(
        ...,
        description="measurement_iso_interval",
        max_length=55,
        title="Measurement Iso Interval",
    )
    measurement_id: UUID = Field(
        ..., description="measurement_id, PRIMARY KEY", title="Measurement Id"
    )
    measurement_datetime: datetime | None = Field(
        None, description="measurement_datetime", title="Measurement Datetime"
    )
    value_as_number: float | None = Field(
        None, description="value_as_number", title="Value As Number"
    )
    range_low: float | None = Field(None, description="range_low", title="Range Low")
    range_high: float | None = Field(None, description="range_high", title="Range High")
    operator_concept_id: int | None = Field(
        None,
        description="operator_concept_id",
        le=2147483647,
        title="Operator Concept Id",
    )
    measurement_time: str | None = Field(
        None, description="measurement_time", max_length=10, title="Measurement Time"
    )
    measurement_source_value: str | None = Field(
        None,
        description="measurement_source_value",
        max_length=50,
        title="Measurement Source Value",
    )
    unit_source_value: str | None = Field(
        None, description="unit_source_value", max_length=50, title="Unit Source Value"
    )
    value_source_value: str | None = Field(
        None,
        description="value_source_value",
        max_length=50,
        title="Value Source Value",
    )
    unit_concept_id: int | None = Field(
        None, description="unit_concept_id", le=2147483647, title="Unit Concept Id"
    )
    value_as_concept_id: int | None = Field(
        None,
        description="value_as_concept_id",
        le=2147483647,
        title="Value As Concept Id",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provided_by_organization_id: UUID | None = Field(
        None,
        description="provided_by_organization_id",
        title="Provided By Organization Id",
    )
    measurement_derived_from_specimen_id: UUID | None = Field(
        None,
        description="measurement_derived_from_specimen_id",
        title="Measurement Derived From Specimen Id",
    )
    measurement_derived_from_measurement_id: UUID | None = Field(
        None,
        description="measurement_derived_from_measurement_id",
        title="Measurement Derived From Measurement Id",
    )
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")


class Observation(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Observations",
        table_name="observation",
        persistable=True,
        keys=create_keys({1: "observation_id"}),
        **_ENTITY_KWARGS,
    )

    obs_event_field_concept_id: int | None = Field(
        None,
        description="obs_event_field_concept_id",
        le=2147483647,
        title="Obs Event Field Concept Id",
    )
    observation_concept_id: int = Field(
        ...,
        description="observation_concept_id",
        le=2147483647,
        title="Observation Concept Id",
    )
    observation_source_concept_id: int = Field(
        ...,
        description="observation_source_concept_id",
        le=2147483647,
        title="Observation Source Concept Id",
    )
    observation_type_concept_id: int = Field(
        ...,
        description="observation_type_concept_id",
        le=2147483647,
        title="Observation Type Concept Id",
    )
    qualifier_concept_id: int | None = Field(
        None,
        description="qualifier_concept_id",
        le=2147483647,
        title="Qualifier Concept Id",
    )
    observation_date: date | None = Field(
        None, description="observation_date", title="Observation Date"
    )
    observation_datetime: datetime | None = Field(
        None, description="observation_datetime", title="Observation Datetime"
    )
    observation_id: UUID = Field(
        ..., description="observation_id, PRIMARY KEY", title="Observation Id"
    )
    observation_event_id: int | None = Field(
        None, description="observation_event_id", title="Observation Event Id"
    )
    observation_iso_interval: str = Field(
        ...,
        description="observation_iso_interval",
        max_length=55,
        title="Observation Iso Interval",
    )
    value_as_datetime: datetime | None = Field(
        None, description="value_as_datetime", title="Value As Datetime"
    )
    value_as_iso_interval: str | None = Field(
        None,
        description="value_as_iso_interval",
        max_length=55,
        title="Value As Iso Interval",
    )
    value_as_number: float | None = Field(
        None, description="value_as_number", title="Value As Number"
    )
    observation_source_value: str | None = Field(
        None,
        description="observation_source_value",
        max_length=50,
        title="Observation Source Value",
    )
    unit_source_value: str | None = Field(
        None, description="unit_source_value", max_length=50, title="Unit Source Value"
    )
    qualifier_source_value: str | None = Field(
        None,
        description="qualifier_source_value",
        max_length=50,
        title="Qualifier Source Value",
    )
    value_as_string: str | None = Field(
        None, description="value_as_string", max_length=60, title="Value As String"
    )
    unit_concept_id: int | None = Field(
        None, description="unit_concept_id", le=2147483647, title="Unit Concept Id"
    )
    value_as_concept_id: int | None = Field(
        None,
        description="value_as_concept_id",
        le=2147483647,
        title="Value As Concept Id",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class Note(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Notes",
        table_name="note",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    encoding_concept_id: int = Field(
        ...,
        description="encoding_concept_id",
        le=2147483647,
        title="Encoding Concept Id",
    )
    language_concept_id: int = Field(
        ...,
        description="language_concept_id",
        le=2147483647,
        title="Language Concept Id",
    )
    note_class_concept_id: int = Field(
        ...,
        description="note_class_concept_id",
        le=2147483647,
        title="Note Class Concept Id",
    )
    note_event_field_concept_id: int | None = Field(
        None,
        description="note_event_field_concept_id",
        le=2147483647,
        title="Note Event Field Concept Id",
    )
    note_type_concept_id: int = Field(
        ...,
        description="note_type_concept_id",
        le=2147483647,
        title="Note Type Concept Id",
    )
    note_date: date = Field(..., description="note_date", title="Note Date")
    note_datetime: datetime | None = Field(
        None, description="note_datetime", title="Note Datetime"
    )
    note_event_id: int | None = Field(
        None, description="note_event_id", title="Note Event Id"
    )
    note_id: UUID = Field(..., description="note_id, PRIMARY KEY", title="Note Id")
    note_title: str | None = Field(
        None, description="note_title", max_length=250, title="Note Title"
    )
    note_source_value: str | None = Field(
        None, description="note_source_value", max_length=50, title="Note Source Value"
    )
    note_text: str = Field(..., description="note_text", title="Note Text")
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    visit_detail_id: UUID | None = Field(
        None, description="visit_detail_id", title="Visit Detail Id"
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class NoteNlp(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="NoteNlps",
        table_name="note_nlp",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    note_nlp_concept_id: int | None = Field(
        None,
        description="note_nlp_concept_id",
        le=2147483647,
        title="Note Nlp Concept Id",
    )
    note_nlp_source_concept_id: int | None = Field(
        None,
        description="note_nlp_source_concept_id",
        le=2147483647,
        title="Note Nlp Source Concept Id",
    )
    nlp_date: date = Field(..., description="nlp_date", title="Nlp Date")
    nlp_datetime: datetime | None = Field(
        None, description="nlp_datetime", title="Nlp Datetime"
    )
    note_nlp_id: UUID = Field(
        ..., description="note_nlp_id, PRIMARY KEY", title="Note Nlp Id"
    )
    note_id: UUID = Field(..., description="note_id", title="Note Id")
    section_concept_id: int | None = Field(
        None,
        description="section_concept_id",
        le=2147483647,
        title="Section Concept Id",
    )
    term_exists: str | None = Field(
        None, description="term_exists", max_length=1, title="Term Exists"
    )
    term_modifiers: str | None = Field(
        None, description="term_modifiers", max_length=2000, title="Term Modifiers"
    )
    snippet: str | None = Field(
        None, description="snippet", max_length=250, title="Snippet"
    )
    lexical_variant: str = Field(
        ..., description="lexical_variant", max_length=250, title="Lexical Variant"
    )
    nlp_system: str | None = Field(
        None, description="nlp_system", max_length=250, title="Nlp System"
    )
    offset: str | None = Field(
        None, description="offset", max_length=50, title="Offset"
    )
    term_temporal: str | None = Field(
        None, description="term_temporal", max_length=50, title="Term Temporal"
    )


class Specimen(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Specimens",
        table_name="specimen",
        persistable=True,
        keys=create_keys({1: "specimen_id"}),
        **_ENTITY_KWARGS,
    )

    anatomic_site_concept_id: int | None = Field(
        None,
        description="anatomic_site_concept_id",
        le=2147483647,
        title="Anatomic Site Concept Id",
    )
    disease_status_concept_id: int | None = Field(
        None,
        description="disease_status_concept_id",
        le=2147483647,
        title="Disease Status Concept Id",
    )
    specimen_concept_id: int = Field(
        ...,
        description="specimen_concept_id",
        le=2147483647,
        title="Specimen Concept Id",
    )
    specimen_type_concept_id: int = Field(
        ...,
        description="specimen_type_concept_id",
        le=2147483647,
        title="Specimen Type Concept Id",
    )
    unit_concept_id: int | None = Field(
        None, description="unit_concept_id", le=2147483647, title="Unit Concept Id"
    )
    specimen_id: UUID = Field(
        ..., description="specimen_id, PRIMARY KEY", title="Specimen Id"
    )
    specimen_date: date | None = Field(
        None, description="specimen_date", title="Specimen Date"
    )
    specimen_datetime: datetime | None = Field(
        None, description="specimen_datetime", title="Specimen Datetime"
    )
    quantity: float | None = Field(None, description="quantity", title="Quantity")
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    specimen_source_id: str | None = Field(
        None,
        description="specimen_source_id",
        max_length=50,
        title="Specimen Source Id",
    )
    specimen_source_value: str | None = Field(
        None,
        description="specimen_source_value",
        max_length=50,
        title="Specimen Source Value",
    )
    unit_source_value: str | None = Field(
        None, description="unit_source_value", max_length=50, title="Unit Source Value"
    )
    anatomic_site_source_value: str | None = Field(
        None,
        description="anatomic_site_source_value",
        max_length=50,
        title="Anatomic Site Source Value",
    )
    disease_status_source_value: str | None = Field(
        None,
        description="disease_status_source_value",
        max_length=50,
        title="Disease Status Source Value",
    )
    specimen_iso_interval: str = Field(
        ...,
        description="specimen_iso_interval",
        max_length=55,
        title="Specimen Iso Interval",
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    derived_from_specimen_id: UUID | None = Field(
        None, description="derived_from_specimen_id", title="Derived From Specimen Id"
    )
    derivation_concept_id: int | None = Field(
        None,
        description="derivation_concept_id",
        le=2147483647,
        title="Derivation Concept Id",
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class FactRelationship(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="FactRelationships",
        table_name="fact_relationship",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    domain_concept_id_1: int = Field(
        ...,
        description="domain_concept_id_1",
        le=2147483647,
        title="Domain Concept Id 1",
    )
    domain_concept_id_2: int = Field(
        ...,
        description="domain_concept_id_2",
        le=2147483647,
        title="Domain Concept Id 2",
    )
    fact_id_1: int = Field(..., description="fact_id_1", title="Fact Id 1")
    fact_id_2: int = Field(..., description="fact_id_2", title="Fact Id 2")
    relationship_concept_id: int = Field(
        ...,
        description="relationship_concept_id",
        le=2147483647,
        title="Relationship Concept Id",
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")


class SurveyConduct(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="SurveyConducts",
        table_name="survey_conduct",
        persistable=True,
        keys=create_keys({1: "survey_conduct_id"}),
        **_ENTITY_KWARGS,
    )

    assisted_concept_id: int = Field(
        ...,
        description="assisted_concept_id",
        le=2147483647,
        title="Assisted Concept Id",
    )
    collection_method_concept_id: int = Field(
        ...,
        description="collection_method_concept_id",
        le=2147483647,
        title="Collection Method Concept Id",
    )
    respondent_type_concept_id: int = Field(
        ...,
        description="respondent_type_concept_id",
        le=2147483647,
        title="Respondent Type Concept Id",
    )
    survey_concept_id: int = Field(
        ..., description="survey_concept_id", le=2147483647, title="Survey Concept Id"
    )
    survey_source_concept_id: int = Field(
        ...,
        description="survey_source_concept_id",
        le=2147483647,
        title="Survey Source Concept Id",
    )
    timing_concept_id: int = Field(
        ..., description="timing_concept_id", le=2147483647, title="Timing Concept Id"
    )
    validated_survey_concept_id: int = Field(
        ...,
        description="validated_survey_concept_id",
        le=2147483647,
        title="Validated Survey Concept Id",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")
    survey_conduct_id: UUID = Field(
        ..., description="survey_conduct_id, PRIMARY KEY", title="Survey Conduct Id"
    )
    survey_end_date: date | None = Field(
        None, description="survey_end_date", title="Survey End Date"
    )
    survey_end_datetime: datetime = Field(
        ..., description="survey_end_datetime", title="Survey End Datetime"
    )
    survey_start_date: date | None = Field(
        None, description="survey_start_date", title="Survey Start Date"
    )
    survey_start_datetime: datetime | None = Field(
        None, description="survey_start_datetime", title="Survey Start Datetime"
    )
    provider_id: UUID | None = Field(
        None, description="provider_id", title="Provider Id"
    )
    respondent_type_source_value: str | None = Field(
        None,
        description="respondent_type_source_value",
        max_length=100,
        title="Respondent Type Source Value",
    )
    timing_source_value: str | None = Field(
        None,
        description="timing_source_value",
        max_length=100,
        title="Timing Source Value",
    )
    collection_method_source_value: str | None = Field(
        None,
        description="collection_method_source_value",
        max_length=100,
        title="Collection Method Source Value",
    )
    survey_source_value: str | None = Field(
        None,
        description="survey_source_value",
        max_length=100,
        title="Survey Source Value",
    )
    survey_source_identifier: str | None = Field(
        None,
        description="survey_source_identifier",
        max_length=100,
        title="Survey Source Identifier",
    )
    validated_survey_source_value: str | None = Field(
        None,
        description="validated_survey_source_value",
        max_length=100,
        title="Validated Survey Source Value",
    )
    survey_version_number: str | None = Field(
        None,
        description="survey_version_number",
        max_length=20,
        title="Survey Version Number",
    )
    assisted_source_value: str | None = Field(
        None,
        description="assisted_source_value",
        max_length=50,
        title="Assisted Source Value",
    )
    response_visit_occurrence_id: UUID | None = Field(
        None,
        description="response_visit_occurrence_id",
        title="Response Visit Occurrence Id",
    )
    visit_occurrence_id: UUID | None = Field(
        None, description="visit_occurrence_id", title="Visit Occurrence Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class Location(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Locations",
        table_name="location",
        persistable=True,
        keys=create_keys({1: "location_id"}),
        **_ENTITY_KWARGS,
    )

    location_id: UUID = Field(
        ..., description="location_id, PRIMARY KEY", title="Location Id"
    )
    latitude: float | None = Field(
        None, description="latitude", ge=-90.0, le=90.0, title="Latitude"
    )
    longitude: float | None = Field(
        None, description="longitude", ge=-180.0, le=180.0, title="Longitude"
    )
    state: str | None = Field(None, description="state", max_length=2, title="State")
    county: str | None = Field(
        None, description="county", max_length=20, title="County"
    )
    address_1: str | None = Field(
        None, description="address_1", max_length=50, title="Address 1"
    )
    address_2: str | None = Field(
        None, description="address_2", max_length=50, title="Address 2"
    )
    city: str | None = Field(None, description="city", max_length=50, title="City")
    location_source_value: str | None = Field(
        None,
        description="location_source_value",
        max_length=50,
        title="Location Source Value",
    )
    zip: str | None = Field(None, description="zip", max_length=9, title="Zip")
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class LocationHistory(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="LocationHistorys",
        table_name="location_history",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    end_date: date | None = Field(None, description="end_date", title="End Date")
    end_iso_interval: str | None = Field(
        None, description="end_iso_interval", max_length=55, title="End Iso Interval"
    )
    entity_id: UUID = Field(..., description="entity_id", title="Entity Id")
    start_date: date = Field(..., description="start_date", title="Start Date")
    start_iso_interval: str = Field(
        ..., description="start_iso_interval", max_length=55, title="Start Iso Interval"
    )
    domain_id: str = Field(
        ..., description="domain_id", max_length=50, title="Domain Id"
    )
    relationship_type_concept_id: int = Field(
        ...,
        description="relationship_type_concept_id",
        le=2147483647,
        title="Relationship Type Concept Id",
    )
    location_id: UUID = Field(..., description="location_id", title="Location Id")
    location_history_id: UUID = Field(
        ..., description="location_history_id, PRIMARY KEY", title="Location History Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class CareSite(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="CareSites",
        table_name="care_site",
        persistable=True,
        keys=create_keys({1: "care_site_id"}),
        **_ENTITY_KWARGS,
    )

    care_site_id: UUID = Field(
        ..., description="care_site_id, PRIMARY KEY", title="Care Site Id"
    )
    care_site_name: str | None = Field(
        None, description="care_site_name", max_length=255, title="Care Site Name"
    )
    care_site_source_value: str | None = Field(
        None,
        description="care_site_source_value",
        max_length=50,
        title="Care Site Source Value",
    )
    place_of_service_source_value: str | None = Field(
        None,
        description="place_of_service_source_value",
        max_length=50,
        title="Place Of Service Source Value",
    )
    place_of_service_concept_id: int = Field(
        ...,
        description="place_of_service_concept_id",
        le=2147483647,
        title="Place Of Service Concept Id",
    )
    location_id: UUID | None = Field(
        None, description="location_id", title="Location Id"
    )
    organization_id: UUID = Field(
        ..., description="organization_id", title="Organization Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class Provider(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Providers",
        table_name="provider",
        persistable=True,
        keys=create_keys({1: "provider_id"}),
        **_ENTITY_KWARGS,
    )

    gender_concept_id: int = Field(
        ..., description="gender_concept_id", le=2147483647, title="Gender Concept Id"
    )
    gender_source_concept_id: int = Field(
        ...,
        description="gender_source_concept_id",
        le=2147483647,
        title="Gender Source Concept Id",
    )
    specialty_concept_id: int = Field(
        ...,
        description="specialty_concept_id",
        le=2147483647,
        title="Specialty Concept Id",
    )
    specialty_source_concept_id: int = Field(
        ...,
        description="specialty_source_concept_id",
        le=2147483647,
        title="Specialty Source Concept Id",
    )
    care_site_id: UUID | None = Field(
        None, description="care_site_id", title="Care Site Id"
    )
    provider_id: UUID = Field(
        ..., description="provider_id, PRIMARY KEY", title="Provider Id"
    )
    year_of_birth: int | None = Field(
        None, description="year_of_birth", title="Year Of Birth"
    )
    npi: str | None = Field(None, description="npi", max_length=20, title="Npi")
    dea: str | None = Field(None, description="dea", max_length=20, title="Dea")
    provider_name: str | None = Field(
        None, description="provider_name", max_length=255, title="Provider Name"
    )
    provider_source_value: str | None = Field(
        None,
        description="provider_source_value",
        max_length=50,
        title="Provider Source Value",
    )
    specialty_source_value: str | None = Field(
        None,
        description="specialty_source_value",
        max_length=50,
        title="Specialty Source Value",
    )
    gender_source_value: str | None = Field(
        None,
        description="gender_source_value",
        max_length=50,
        title="Gender Source Value",
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class PayerPlanPeriod(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="PayerPlanPeriods",
        table_name="payer_plan_period",
        persistable=True,
        keys=create_keys({1: "payer_plan_period_id"}),
        **_ENTITY_KWARGS,
    )

    contract_concept_id: int = Field(
        ...,
        description="contract_concept_id",
        le=2147483647,
        title="Contract Concept Id",
    )
    payer_concept_id: int = Field(
        ..., description="payer_concept_id", le=2147483647, title="Payer Concept Id"
    )
    payer_source_concept_id: int = Field(
        ...,
        description="payer_source_concept_id",
        le=2147483647,
        title="Payer Source Concept Id",
    )
    plan_concept_id: int = Field(
        ..., description="plan_concept_id", le=2147483647, title="Plan Concept Id"
    )
    plan_source_concept_id: int = Field(
        ...,
        description="plan_source_concept_id",
        le=2147483647,
        title="Plan Source Concept Id",
    )
    sponsor_concept_id: int = Field(
        ..., description="sponsor_concept_id", le=2147483647, title="Sponsor Concept Id"
    )
    sponsor_source_concept_id: int | None = Field(
        None,
        description="sponsor_source_concept_id",
        le=2147483647,
        title="Sponsor Source Concept Id",
    )
    stop_reason_concept_id: int | None = Field(
        None,
        description="stop_reason_concept_id",
        le=2147483647,
        title="Stop Reason Concept Id",
    )
    stop_reason_source_concept_id: int | None = Field(
        None,
        description="stop_reason_source_concept_id",
        le=2147483647,
        title="Stop Reason Source Concept Id",
    )
    payer_plan_period_end_date: date = Field(
        ...,
        description="payer_plan_period_end_date",
        title="Payer Plan Period End Date",
    )
    payer_plan_period_start_date: date = Field(
        ...,
        description="payer_plan_period_start_date",
        title="Payer Plan Period Start Date",
    )
    payer_source_value: str | None = Field(
        None,
        description="payer_source_value",
        max_length=50,
        title="Payer Source Value",
    )
    plan_source_value: str | None = Field(
        None, description="plan_source_value", max_length=50, title="Plan Source Value"
    )
    contract_source_value: str = Field(
        ...,
        description="contract_source_value",
        max_length=50,
        title="Contract Source Value",
    )
    sponsor_source_value: str | None = Field(
        None,
        description="sponsor_source_value",
        max_length=50,
        title="Sponsor Source Value",
    )
    family_source_value: str | None = Field(
        None,
        description="family_source_value",
        max_length=50,
        title="Family Source Value",
    )
    stop_reason_source_value: str | None = Field(
        None,
        description="stop_reason_source_value",
        max_length=50,
        title="Stop Reason Source Value",
    )
    contract_person_id: UUID | None = Field(
        None, description="contract_person_id", title="Contract Person Id"
    )
    contract_source_concept_id: int = Field(
        ...,
        description="contract_source_concept_id",
        le=2147483647,
        title="Contract Source Concept Id",
    )
    payer_plan_period_id: UUID = Field(
        ...,
        description="payer_plan_period_id, PRIMARY KEY",
        title="Payer Plan Period Id",
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")


class Cost(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Costs",
        table_name="cost",
        persistable=True,
        keys=create_keys({1: "cost_id"}),
        **_ENTITY_KWARGS,
    )

    cost_concept_id: int | None = Field(
        None, description="cost_concept_id", le=2147483647, title="Cost Concept Id"
    )
    billed_date: date | None = Field(
        None, description="billed_date", title="Billed Date"
    )
    cost_event_field_concept_id: int = Field(
        ...,
        description="cost_event_field_concept_id",
        le=2147483647,
        title="Cost Event Field Concept Id",
    )
    cost_source_concept_id: int | None = Field(
        None,
        description="cost_source_concept_id",
        le=2147483647,
        title="Cost Source Concept Id",
    )
    cost_type_concept_id: int | None = Field(
        None,
        description="cost_type_concept_id",
        le=2147483647,
        title="Cost Type Concept Id",
    )
    cost_id: UUID = Field(..., description="cost_id, PRIMARY KEY", title="Cost Id")
    person_id: int = Field(..., description="person_id", title="Person Id")
    cost_event_id: int = Field(..., description="cost_event_id", title="Cost Event Id")
    payer_plan_period_id: int | None = Field(
        None, description="payer_plan_period_id", title="Payer Plan Period Id"
    )
    incurred_date: date | None = Field(
        None, description="incurred_date", title="Incurred Date"
    )
    paid_date: date | None = Field(None, description="paid_date", title="Paid Date")
    cost: float | None = Field(None, description="cost", title="Cost")
    currency_concept_id: int | None = Field(
        None,
        description="currency_concept_id",
        le=2147483647,
        title="Currency Concept Id",
    )
    cost_source_value: str | None = Field(
        None, description="cost_source_value", max_length=50, title="Cost Source Value"
    )
    revenue_code_source_value: str | None = Field(
        None,
        description="revenue_code_source_value",
        max_length=50,
        title="Revenue Code Source Value",
    )
    drg_source_value: str | None = Field(
        None, description="drg_source_value", max_length=50, title="Drg Source Value"
    )
    drg_concept_id: int | None = Field(
        None, description="drg_concept_id", le=2147483647, title="Drg Concept Id"
    )
    revenue_code_concept_id: int | None = Field(
        None,
        description="revenue_code_concept_id",
        le=2147483647,
        title="Revenue Code Concept Id",
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class DrugEra(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="DrugEras",
        table_name="drug_era",
        persistable=True,
        keys=create_keys({1: "drug_era_id"}),
        **_ENTITY_KWARGS,
    )

    drug_concept_id: int = Field(
        ..., description="drug_concept_id", le=2147483647, title="Drug Concept Id"
    )
    drug_era_end_datetime: datetime = Field(
        ..., description="drug_era_end_datetime", title="Drug Era End Datetime"
    )
    drug_era_id: UUID = Field(
        ..., description="drug_era_id, PRIMARY KEY", title="Drug Era Id"
    )
    drug_era_end_iso_interval: str = Field(
        ...,
        description="drug_era_end_iso_interval",
        max_length=55,
        title="Drug Era End Iso Interval",
    )
    drug_era_start_datetime: datetime = Field(
        ..., description="drug_era_start_datetime", title="Drug Era Start Datetime"
    )
    drug_era_start_iso_interval: str = Field(
        ...,
        description="drug_era_start_iso_interval",
        max_length=55,
        title="Drug Era Start Iso Interval",
    )
    drug_exposure_count: int | None = Field(
        None, description="drug_exposure_count", title="Drug Exposure Count"
    )
    gap_days: int | None = Field(None, description="gap_days", title="Gap Days")
    person_id: UUID = Field(..., description="person_id", title="Person Id")


class DoseEra(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="DoseEras",
        table_name="dose_era",
        persistable=True,
        keys=create_keys({1: "dose_era_id"}),
        **_ENTITY_KWARGS,
    )

    dose_era_end_datetime: datetime = Field(
        ..., description="dose_era_end_datetime", title="Dose Era End Datetime"
    )
    drug_concept_id: int = Field(
        ..., description="drug_concept_id", le=2147483647, title="Drug Concept Id"
    )
    dose_era_id: UUID = Field(
        ..., description="dose_era_id, PRIMARY KEY", title="Dose Era Id"
    )
    dose_era_start_datetime: datetime = Field(
        ..., description="dose_era_start_datetime", title="Dose Era Start Datetime"
    )
    dose_value: float = Field(..., description="dose_value", title="Dose Value")
    unit_concept_id: int = Field(
        ..., description="unit_concept_id", le=2147483647, title="Unit Concept Id"
    )
    person_id: UUID = Field(..., description="person_id", title="Person Id")


class ConditionEra(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConditionEras",
        table_name="condition_era",
        persistable=True,
        keys=create_keys({1: "condition_era_id"}),
        **_ENTITY_KWARGS,
    )

    condition_era_end_datetime: datetime = Field(
        ...,
        description="condition_era_end_datetime",
        title="Condition Era End Datetime",
    )
    condition_concept_id: int = Field(
        ...,
        description="condition_concept_id",
        le=2147483647,
        title="Condition Concept Id",
    )
    condition_era_id: UUID = Field(
        ..., description="condition_era_id, PRIMARY KEY", title="Condition Era Id"
    )
    person_id: int = Field(..., description="person_id", title="Person Id")
    condition_era_start_datetime: datetime = Field(
        ...,
        description="condition_era_start_datetime",
        title="Condition Era Start Datetime",
    )
    condition_occurrence_count: int | None = Field(
        None,
        description="condition_occurrence_count",
        title="Condition Occurrence Count",
    )


class Metadata(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Metadatas",
        table_name="metadata",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    metadata_concept_id: int = Field(
        ...,
        description="metadata_concept_id",
        le=2147483647,
        title="Metadata Concept Id",
    )
    metadata_type_concept_id: int = Field(
        ...,
        description="metadata_type_concept_id",
        le=2147483647,
        title="Metadata Type Concept Id",
    )
    metadata_date: date | None = Field(
        None, description="metadata_date", title="Metadata Date"
    )
    metadata_datetime: datetime | None = Field(
        None, description="metadata_datetime", title="Metadata Datetime"
    )
    name: str = Field(..., description="name", max_length=250, title="Name")
    value_as_string: str | None = Field(
        None, description="value_as_string", max_length=250, title="Value As String"
    )
    value_as_concept_id: int | None = Field(
        None,
        description="value_as_concept_id",
        le=2147483647,
        title="Value As Concept Id",
    )


class CdmSource(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="CdmSources",
        table_name="cdm_source",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    cdm_release_date: date | None = Field(
        None, description="cdm_release_date", title="Cdm Release Date"
    )
    source_release_date: date | None = Field(
        None, description="source_release_date", title="Source Release Date"
    )
    cdm_version: str | None = Field(
        None, description="cdm_version", max_length=10, title="Cdm Version"
    )
    vocabulary_version: str | None = Field(
        None,
        description="vocabulary_version",
        max_length=20,
        title="Vocabulary Version",
    )
    cdm_source_abbreviation: str | None = Field(
        None,
        description="cdm_source_abbreviation",
        max_length=25,
        title="Cdm Source Abbreviation",
    )
    cdm_source_name: str = Field(
        ..., description="cdm_source_name", max_length=255, title="Cdm Source Name"
    )
    cdm_holder: str | None = Field(
        None, description="cdm_holder", max_length=255, title="Cdm Holder"
    )
    source_documentation_reference: str | None = Field(
        None,
        description="source_documentation_reference",
        max_length=255,
        title="Source Documentation Reference",
    )
    cdm_etl_reference: str | None = Field(
        None, description="cdm_etl_reference", max_length=255, title="Cdm Etl Reference"
    )
    source_description: str | None = Field(
        None, description="source_description", title="Source Description"
    )


class Concept(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Concepts",
        table_name="concept",
        persistable=True,
        keys=create_keys({1: "concept_id"}),
        **_ENTITY_KWARGS,
    )

    valid_end_date: date = Field(
        ..., description="valid_end_date", title="Valid End Date"
    )
    valid_start_date: date = Field(
        ..., description="valid_start_date", title="Valid Start Date"
    )
    concept_id: int = Field(
        ..., description="concept_id, PRIMARY KEY", title="Concept Id"
    )
    standard_concept: str | None = Field(
        None, description="standard_concept", max_length=1, title="Standard Concept"
    )
    invalid_reason: str | None = Field(
        None, description="invalid_reason", max_length=1, title="Invalid Reason"
    )
    concept_name: str = Field(
        ..., description="concept_name", max_length=255, title="Concept Name"
    )
    concept_code: str = Field(
        ..., description="concept_code", max_length=50, title="Concept Code"
    )
    concept_class_id: str = Field(
        ..., description="concept_class_id", max_length=20, title="Concept Class Id"
    )
    domain_id: str = Field(
        ..., description="domain_id", max_length=20, title="Domain Id"
    )
    vocabulary_id: str = Field(
        ..., description="vocabulary_id", max_length=20, title="Vocabulary Id"
    )


class Vocabulary(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Vocabularys",
        table_name="vocabulary",
        persistable=True,
        keys=create_keys({1: "vocabulary_id"}),
        **_ENTITY_KWARGS,
    )

    vocabulary_concept_id: int = Field(
        ...,
        description="vocabulary_concept_id",
        le=2147483647,
        title="Vocabulary Concept Id",
    )
    vocabulary_id: str = Field(
        ...,
        description="vocabulary_id, PRIMARY KEY",
        max_length=20,
        title="Vocabulary Id",
    )
    vocabulary_name: str = Field(
        ..., description="vocabulary_name", max_length=255, title="Vocabulary Name"
    )
    vocabulary_reference: str = Field(
        ...,
        description="vocabulary_reference",
        max_length=255,
        title="Vocabulary Reference",
    )
    vocabulary_version: str | None = Field(
        None,
        description="vocabulary_version",
        max_length=255,
        title="Vocabulary Version",
    )


class Domain(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Domains",
        table_name="domain",
        persistable=True,
        keys=create_keys({1: "domain_id"}),
        **_ENTITY_KWARGS,
    )

    domain_concept_id: int = Field(
        ..., description="domain_concept_id", le=2147483647, title="Domain Concept Id"
    )
    domain_id: str = Field(
        ..., description="domain_id, PRIMARY KEY", max_length=20, title="Domain Id"
    )
    domain_name: str = Field(
        ..., description="domain_name", max_length=255, title="Domain Name"
    )


class ConceptClass(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConceptClasss",
        table_name="concept_class",
        persistable=True,
        keys=create_keys({1: "concept_class_id"}),
        **_ENTITY_KWARGS,
    )

    concept_class_concept_id: int = Field(
        ...,
        description="concept_class_concept_id",
        le=2147483647,
        title="Concept Class Concept Id",
    )
    concept_class_id: str = Field(
        ...,
        description="concept_class_id, PRIMARY KEY",
        max_length=20,
        title="Concept Class Id",
    )
    concept_class_name: str = Field(
        ...,
        description="concept_class_name",
        max_length=255,
        title="Concept Class Name",
    )


class ConceptRelationship(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConceptRelationships",
        table_name="concept_relationship",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    valid_end_date: date = Field(
        ..., description="valid_end_date", title="Valid End Date"
    )
    valid_start_date: date = Field(
        ..., description="valid_start_date", title="Valid Start Date"
    )
    concept_id_1: int = Field(
        ..., description="concept_id_1", le=2147483647, title="Concept Id 1"
    )
    invalid_reason: str | None = Field(
        None, description="invalid_reason", max_length=1, title="Invalid Reason"
    )
    concept_id_2: int = Field(
        ..., description="concept_id_2", le=2147483647, title="Concept Id 2"
    )
    relationship_id: str = Field(
        ..., description="relationship_id", max_length=20, title="Relationship Id"
    )


class Relationship(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Relationships",
        table_name="relationship",
        persistable=True,
        keys=create_keys({1: "relationship_id"}),
        **_ENTITY_KWARGS,
    )

    relationship_concept_id: int = Field(
        ...,
        description="relationship_concept_id",
        le=2147483647,
        title="Relationship Concept Id",
    )
    is_hierarchical: str = Field(
        ..., description="is_hierarchical", max_length=1, title="Is Hierarchical"
    )
    defines_ancestry: str = Field(
        ..., description="defines_ancestry", max_length=1, title="Defines Ancestry"
    )
    relationship_id: str = Field(
        ...,
        description="relationship_id, PRIMARY KEY",
        max_length=20,
        title="Relationship Id",
    )
    reverse_relationship_id: str = Field(
        ...,
        description="reverse_relationship_id",
        max_length=20,
        title="Reverse Relationship Id",
    )
    relationship_name: str = Field(
        ..., description="relationship_name", max_length=255, title="Relationship Name"
    )


class ConceptSynonym(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConceptSynonyms",
        table_name="concept_synonym",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    concept_id: int = Field(
        ..., description="concept_id", le=2147483647, title="Concept Id"
    )
    concept_synonym_name: str = Field(
        ...,
        description="concept_synonym_name",
        max_length=1000,
        title="Concept Synonym Name",
    )
    language_concept_id: int = Field(
        ...,
        description="language_concept_id",
        le=2147483647,
        title="Language Concept Id",
    )


class ConceptAncestor(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="ConceptAncestors",
        table_name="concept_ancestor",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    ancestor_concept_id: int = Field(
        ...,
        description="ancestor_concept_id",
        le=2147483647,
        title="Ancestor Concept Id",
    )
    min_levels_of_separation: int = Field(
        ..., description="min_levels_of_separation", title="Min Levels Of Separation"
    )
    max_levels_of_separation: int = Field(
        ..., description="max_levels_of_separation", title="Max Levels Of Separation"
    )
    descendant_concept_id: int = Field(
        ...,
        description="descendant_concept_id",
        le=2147483647,
        title="Descendant Concept Id",
    )


class SourceToConceptMap(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="SourceToConceptMaps",
        table_name="source_to_concept_map",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    source_concept_id: int = Field(
        ..., description="source_concept_id", le=2147483647, title="Source Concept Id"
    )
    target_concept_id: int = Field(
        ..., description="target_concept_id", le=2147483647, title="Target Concept Id"
    )
    valid_end_date: date = Field(
        ..., description="valid_end_date", title="Valid End Date"
    )
    valid_start_date: date = Field(
        ..., description="valid_start_date", title="Valid Start Date"
    )
    invalid_reason: str | None = Field(
        None, description="invalid_reason", max_length=1, title="Invalid Reason"
    )
    source_vocabulary_id: str = Field(
        ...,
        description="source_vocabulary_id",
        max_length=20,
        title="Source Vocabulary Id",
    )
    source_code_description: str | None = Field(
        None,
        description="source_code_description",
        max_length=255,
        title="Source Code Description",
    )
    source_code: str = Field(
        ..., description="source_code", max_length=50, title="Source Code"
    )
    target_vocabulary_id: str = Field(
        ...,
        description="target_vocabulary_id",
        max_length=20,
        title="Target Vocabulary Id",
    )


class DrugStrength(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="DrugStrengths",
        table_name="drug_strength",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    amount_unit_concept_id: int | None = Field(
        None,
        description="amount_unit_concept_id",
        le=2147483647,
        title="Amount Unit Concept Id",
    )
    denominator_unit_concept_id: int | None = Field(
        None,
        description="denominator_unit_concept_id",
        le=2147483647,
        title="Denominator Unit Concept Id",
    )
    drug_concept_id: int = Field(
        ..., description="drug_concept_id", le=2147483647, title="Drug Concept Id"
    )
    valid_end_date: date = Field(
        ..., description="valid_end_date", title="Valid End Date"
    )
    valid_start_date: date = Field(
        ..., description="valid_start_date", title="Valid Start Date"
    )
    ingredient_concept_id: int = Field(
        ...,
        description="ingredient_concept_id",
        le=2147483647,
        title="Ingredient Concept Id",
    )
    amount_value: float | None = Field(
        None, description="amount_value", title="Amount Value"
    )
    numerator_value: float | None = Field(
        None, description="numerator_value", title="Numerator Value"
    )
    denominator_value: float | None = Field(
        None, description="denominator_value", title="Denominator Value"
    )
    box_size: int | None = Field(None, description="box_size", title="Box Size")
    invalid_reason: str | None = Field(
        None, description="invalid_reason", max_length=1, title="Invalid Reason"
    )
    numerator_unit_concept_id: int | None = Field(
        None,
        description="numerator_unit_concept_id",
        le=2147483647,
        title="Numerator Unit Concept Id",
    )


class Cohort(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Cohorts",
        table_name="cohort",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    cohort_end_date: date = Field(
        ..., description="cohort_end_date", title="Cohort End Date"
    )
    cohort_start_date: date = Field(
        ..., description="cohort_start_date", title="Cohort Start Date"
    )
    cohort_definition_id: UUID = Field(
        ..., description="cohort_definition_id", title="Cohort Definition Id"
    )
    subject_id: UUID = Field(..., description="subject_id", title="Subject Id")


class CohortDefinition(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="CohortDefinitions",
        table_name="cohort_definition",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    cohort_initiation_date: date | None = Field(
        None, description="cohort_initiation_date", title="Cohort Initiation Date"
    )
    cohort_definition_name: str = Field(
        ...,
        description="cohort_definition_name",
        max_length=255,
        title="Cohort Definition Name",
    )
    cohort_definition_description: str | None = Field(
        None,
        description="cohort_definition_description",
        title="Cohort Definition Description",
    )
    cohort_definition_syntax: str | None = Field(
        None, description="cohort_definition_syntax", title="Cohort Definition Syntax"
    )
    definition_type_concept_id: int = Field(
        ...,
        description="definition_type_concept_id",
        le=2147483647,
        title="Definition Type Concept Id",
    )
    subject_concept_id: int = Field(
        ..., description="subject_concept_id", le=2147483647, title="Subject Concept Id"
    )
    cohort_definition_id: UUID = Field(
        ..., description="cohort_definition_id", title="Cohort Definition Id"
    )


class Provenance(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Provenances",
        table_name="provenance",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    provenance_id: UUID = Field(
        ..., description="provenance_id, PRIMARY KEY", title="Provenance Id"
    )
    source_id: UUID = Field(..., description="source_id", title="Source Id")
    etl_id: UUID = Field(..., description="etl_id", title="Etl Id")
    etl_start_datetime: datetime = Field(
        ..., description="etl_start_datetime", title="Etl Start Datetime"
    )
    etl_end_datetime: datetime = Field(
        ..., description="etl_end_datetime", title="Etl End Datetime"
    )


class Etl(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Etls",
        table_name="etl",
        persistable=True,
        keys=create_keys({1: "etl_id"}),
        **_ENTITY_KWARGS,
    )

    etl_id: UUID = Field(..., description="etl_id, PRIMARY KEY", title="Etl Id")
    etl_desc: str = Field(..., description="etl_desc", title="Etl Desc")
    etl_abbrev: str = Field(
        ..., description="etl_abbrev", max_length=32, title="Etl Abbrev"
    )
    etl_version: str = Field(
        ..., description="etl_version", max_length=32, title="Etl Version"
    )
    etl_release_datetime: datetime = Field(
        ..., description="etl_release_datetime", title="Etl Release Datetime"
    )


class Source(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Sources",
        table_name="source",
        persistable=True,
        keys=create_keys({1: "source_id"}),
        **_ENTITY_KWARGS,
    )

    source_id: UUID = Field(
        ..., description="source_id, PRIMARY KEY", title="Source Id"
    )
    source_desc: str = Field(..., description="source_desc", title="Source Desc")
    source_abbrev: str = Field(
        ..., description="source_abbrev", max_length=32, title="Source Abbrev"
    )


class Organization(Model):

    ENTITY: ClassVar = Entity(
        snake_case_plural_name="Organizations",
        table_name="organization",
        persistable=True,
        keys=create_keys({}),
        **_ENTITY_KWARGS,
    )

    organization_id: UUID = Field(
        ..., description="organization_id, PRIMARY KEY", title="Organization Id"
    )
    provenance_id: UUID = Field(..., description="provenance_id", title="Provenance Id")
    source_traceback: str = Field(
        ..., description="source_traceback", title="Source Traceback"
    )


class Subject(Model):

    id: int
    person: Person | None = Field(
        default=None,
    )
    specimen_records: list[Specimen]
    observation_records: list[Observation]
    measurement_records: list[Measurement]
    drug_exposure_records: list[DrugExposure]
    location_history_records: list[LocationHistory]


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import datetime
from typing import Type
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, relationship

from gen_epix.casedb.domain import enum, model
from gen_epix.casedb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.casedb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: Type = sa.orm.declarative_base(name=enum.ServiceType.ORGANIZATION.value)


class Organization(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Organization)

    name: Mapped[str] = create_mapped_column(model.Organization, "name")
    legal_entity_code: Mapped[str] = create_mapped_column(
        model.Organization, "legal_entity_code"
    )
    legal_region_id: Mapped[UUID | None] = create_mapped_column(
        model.Organization, "legal_region_id"
    )


class User(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.User)

    email: Mapped[str] = create_mapped_column(model.User, "email")
    name: Mapped[str | None] = create_mapped_column(model.User, "name")
    is_active: Mapped[bool] = create_mapped_column(model.User, "is_active")
    roles: Mapped[set[enum.Role]] = create_mapped_column(model.User, "roles")
    organization_id: Mapped[UUID] = create_mapped_column(model.User, "organization_id")

    organization: Mapped[Organization] = relationship(
        Organization, foreign_keys=[organization_id]
    )


class OrganizationSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationSet)

    name: Mapped[str] = create_mapped_column(model.OrganizationSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        model.OrganizationSet, "description"
    )


class OrganizationSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.OrganizationSetMember)

    organization_set_id: Mapped[UUID] = create_mapped_column(
        model.OrganizationSetMember, "organization_set_id"
    )
    organization_id: Mapped[UUID] = create_mapped_column(
        model.OrganizationSetMember, "organization_id"
    )

    organization_set: Mapped[OrganizationSet] = relationship(
        OrganizationSet, foreign_keys=[organization_set_id]
    )
    organization: Mapped[Organization] = relationship(
        Organization, foreign_keys=[organization_id]
    )


class Site(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Site)

    organization_id: Mapped[UUID] = create_mapped_column(model.Site, "organization_id")
    name: Mapped[str] = create_mapped_column(model.Site, "name")
    location_region_id: Mapped[UUID] = create_mapped_column(
        model.Site, "location_region_id"
    )

    organization: Mapped[Organization] = relationship(
        Organization, foreign_keys=[organization_id]
    )


class Contact(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Contact)

    site_id: Mapped[UUID | None] = create_mapped_column(model.Contact, "site_id")
    name: Mapped[str] = create_mapped_column(model.Contact, "name")
    email: Mapped[str | None] = create_mapped_column(model.Contact, "email")
    phone: Mapped[str | None] = create_mapped_column(model.Contact, "phone")

    site: Mapped[Site] = relationship(Site, foreign_keys=[site_id])


class IdentifierIssuer(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.IdentifierIssuer)

    name: Mapped[str] = create_mapped_column(model.IdentifierIssuer, "name")


class DataCollection(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollection)

    name: Mapped[str] = create_mapped_column(model.DataCollection, "name")
    description: Mapped[str | None] = create_mapped_column(
        model.DataCollection, "description"
    )


class DataCollectionSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSet)

    name: Mapped[str] = create_mapped_column(model.DataCollectionSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        model.DataCollectionSet, "description"
    )


class DataCollectionSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSetMember)

    data_collection_set_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionSetMember, "data_collection_set_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionSetMember, "data_collection_id"
    )

    data_collection_set: Mapped[DataCollectionSet] = relationship(
        DataCollectionSet, foreign_keys=[data_collection_set_id]
    )
    data_collection: Mapped[DataCollection] = relationship(
        DataCollection, foreign_keys=[data_collection_id]
    )


class DataCollectionRelation(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionRelation)

    from_data_collection_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionRelation, "from_data_collection_id"
    )
    to_data_collection_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionRelation, "to_data_collection_id"
    )
    share_case: Mapped[bool] = create_mapped_column(
        model.DataCollectionRelation, "share_case"
    )

    from_data_collection: Mapped[DataCollection] = relationship(
        DataCollection, foreign_keys=[from_data_collection_id]
    )
    to_data_collection: Mapped[DataCollection] = relationship(
        DataCollection, foreign_keys=[to_data_collection_id]
    )


class UserInvitation(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.UserInvitation)

    email: Mapped[str] = create_mapped_column(model.UserInvitation, "email")
    token: Mapped[str] = create_mapped_column(model.UserInvitation, "token")
    expires_at: Mapped[datetime.datetime] = create_mapped_column(
        model.UserInvitation, "expires_at"
    )
    roles: Mapped[set[enum.Role]] = create_mapped_column(model.UserInvitation, "roles")
    invited_by_user_id: Mapped[UUID] = create_mapped_column(
        model.UserInvitation, "invited_by_user_id"
    )
    organization_id: Mapped[UUID] = create_mapped_column(
        model.UserInvitation, "organization_id"
    )

    invited_by_user: Mapped[User] = relationship(
        User, foreign_keys=[invited_by_user_id]
    )
    organization: Mapped[Organization] = relationship(
        Organization, foreign_keys=[organization_id]
    )



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar
from uuid import UUID

from gen_epix.fastapp.services import auth
from gen_epix.omopdb.domain import DOMAIN, enum
from gen_epix.omopdb.domain.command.base import Command


# auth
class GetIdentityProvidersCommand(auth.GetIdentityProvidersCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.AUTH


# organization
class RetrieveCompleteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION


class InviteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    email: str
    roles: set[enum.Role]


class RegisterInvitedUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    token: str


class UpdateUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    tgt_user_id: UUID
    is_active: bool | None
    roles: set[enum.Role] | None
    data_collection_ids: set[UUID] | None
    organization_id: UUID | None


# system
class RetrieveOutagesCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SYSTEM
    pass


DOMAIN.register_locals(locals())



from gen_epix.fastapp.model import Policy
from gen_epix.seqdb.domain.service.organization import BaseOrganizationService
from gen_epix.seqdb.domain.service.rbac import BaseRbacService


class BaseUpdateUserPolicy(Policy):
    def __init__(
        self,
        rbac_service: BaseRbacService,
        organization_service: BaseOrganizationService,
        **kwargs: dict,
    ):
        self.rbac_service = rbac_service
        self.organization_service = organization_service
        self.props = kwargs



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


import datetime
from typing import Type
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.orm import Mapped

from gen_epix.omopdb.domain import enum, model
from gen_epix.omopdb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.omopdb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: Type = sa.orm.declarative_base(name=enum.ServiceType.ORGANIZATION.value)


class User(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.User)

    email: Mapped[str] = create_mapped_column(model.User, "email")
    name: Mapped[str | None] = create_mapped_column(model.User, "name")
    is_active: Mapped[bool] = create_mapped_column(model.User, "is_active")
    roles: Mapped[set[enum.Role]] = create_mapped_column(model.User, "roles")
    data_collection_ids: Mapped[set[UUID]] = create_mapped_column(
        model.User, "data_collection_ids"
    )


class UserInvitation(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.UserInvitation)

    email: Mapped[str] = create_mapped_column(model.UserInvitation, "email")
    token: Mapped[str] = create_mapped_column(model.UserInvitation, "token")
    expires_at: Mapped[datetime.datetime] = create_mapped_column(
        model.UserInvitation, "expires_at"
    )
    roles: Mapped[set[enum.Role]] = create_mapped_column(model.UserInvitation, "roles")
    data_collection_ids: Mapped[set[UUID]] = create_mapped_column(
        model.UserInvitation, "data_collection_ids"
    )
    invited_by_user_id: Mapped[UUID] = create_mapped_column(
        model.UserInvitation, "invited_by_user_id"
    )


class IdentifierIssuer(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.IdentifierIssuer)

    name: Mapped[str] = create_mapped_column(model.IdentifierIssuer, "name")


class DataCollection(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollection)

    name: Mapped[str] = create_mapped_column(model.DataCollection, "name")
    description: Mapped[str | None] = create_mapped_column(
        model.DataCollection, "description"
    )


class DataCollectionSet(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSet)

    name: Mapped[str] = create_mapped_column(model.DataCollectionSet, "name")
    description: Mapped[str | None] = create_mapped_column(
        model.DataCollectionSet, "description"
    )


class DataCollectionSetMember(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.DataCollectionSetMember)

    data_collection_set_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionSetMember, "data_collection_set_id"
    )
    data_collection_id: Mapped[UUID] = create_mapped_column(
        model.DataCollectionSetMember, "data_collection_id"
    )



from gen_epix.casedb.domain.repository.abac import BaseAbacRepository
from gen_epix.fastapp.repositories import DictRepository


class AbacDictRepository(DictRepository, BaseAbacRepository):
    pass



import datetime
from typing import Type

import sqlalchemy as sa
from sqlalchemy.orm import Mapped

from gen_epix.seqdb.domain import enum, model
from gen_epix.seqdb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.seqdb.repositories.sa_model.util import (
    create_mapped_column,
    create_table_args,
)

Base: Type = sa.orm.declarative_base(name=enum.ServiceType.SYSTEM.value)


class Outage(Base, RowMetadataMixin):
    __tablename__, __table_args__ = create_table_args(model.Outage)

    description: Mapped[str | None] = create_mapped_column(model.Outage, "description")
    active_from: Mapped[datetime.datetime | None] = create_mapped_column(
        model.Outage, "active_from"
    )
    active_to: Mapped[datetime.datetime | None] = create_mapped_column(
        model.Outage, "active_to"
    )
    visible_from: Mapped[datetime.datetime | None] = create_mapped_column(
        model.Outage, "visible_from"
    )
    visible_to: Mapped[datetime.datetime | None] = create_mapped_column(
        model.Outage, "visible_to"
    )
    is_active: Mapped[bool | None] = create_mapped_column(model.Outage, "is_active")
    is_visible: Mapped[bool | None] = create_mapped_column(model.Outage, "is_visible")



from typing import Any, Type

import sqlalchemy as sa
from sqlalchemy.orm import MappedColumn, mapped_column

from gen_epix.fastapp import Entity
from gen_epix.fastapp.repositories import get_pydantic_field_sa_type
from gen_epix.seqdb.domain.model import Model
from gen_epix.seqdb.domain.model.base import Model


def create_table_args(
    model_class: Type[Model],
    field_name_map: dict[str, str] | None = None,
    **kwargs: dict,
) -> tuple:
    entity: Entity = model_class.ENTITY  # type: ignore[attr-defined]
    uq_constraints = []
    for field_names in entity.get_keys_field_names():
        sa_field_names = (
            [field_name_map.get(x, x) for x in field_names]
            if field_name_map
            else field_names
        )
        sa_field_name_str = "_".join(sa_field_names)
        uq_constraints.append(
            sa.UniqueConstraint(
                *sa_field_names,
                name=f"uq_{entity.table_name}_{sa_field_name_str}",
                **kwargs,
            )
        )
    if entity.schema_name:
        return entity.table_name, tuple(
            [*uq_constraints, {"schema": entity.schema_name}]
        )
    return entity.table_name, tuple([*uq_constraints])


def create_mapped_column(
    model_class: Type[Model],
    field_name: str,
    field_name_map: dict[str, str] | None = None,
    **kwargs: dict,
) -> MappedColumn[Any]:
    entity: Entity = model_class.ENTITY  # type: ignore[attr-defined]
    fieldinfo = model_class.model_fields[field_name]
    sa_type = get_pydantic_field_sa_type(fieldinfo)
    nullable = kwargs.get("nullable", fieldinfo.is_required() is False)
    doc = kwargs.pop("doc", fieldinfo.description)
    link_entity = entity.get_link_entity(field_name)
    if link_entity and link_entity.service_type != entity.service_type:
        # Create foreign keys only within the same service
        link_entity = None
    ondelete = kwargs.pop("ondelete", None)
    onupdate = kwargs.pop("onupdate", None)
    sa_field_name = (
        field_name_map[model_class][field_name] if field_name_map else field_name
    )
    fk_name = kwargs.pop("fk_name", f"fk_{entity.table_name}_{sa_field_name}")
    if link_entity:
        link_model_class = link_entity.model_class
        link_sa_id_field_name = (
            field_name_map[link_model_class][link_entity.id_field_name]
            if field_name_map
            else link_entity.id_field_name
        )
        ref_column_name = (
            f"{link_entity.schema_name}.{link_entity.table_name}.{link_sa_id_field_name}"
            if link_entity.schema_name
            else f"{link_entity.table_name}.{link_sa_id_field_name}"
        )
        return mapped_column(
            sa_type,
            sa.ForeignKey(
                ref_column_name, ondelete=ondelete, onupdate=onupdate, name=fk_name
            ),
            nullable=nullable,
            doc=doc,
            **kwargs,
        )
    return mapped_column(
        sa_type,
        nullable=nullable,
        primary_key=entity.id_field_name == field_name,
        doc=doc,
        **kwargs,
    )



# pylint: disable=useless-import-alias
from gen_epix.seqdb.domain.repository.organization import (
    BaseOrganizationRepository as BaseOrganizationRepository,
)
from gen_epix.seqdb.domain.repository.seq import BaseSeqRepository as BaseSeqRepository
from gen_epix.seqdb.domain.repository.system import (
    BaseSystemRepository as BaseSystemRepository,
)



from gen_epix.fastapp import CrudOperation, EventTiming
from gen_epix.omopdb.domain import command, model
from gen_epix.omopdb.domain.service import BaseSystemService
from gen_epix.omopdb.policies.has_system_outage_policy import HasSystemOutagePolicy


class SystemService(BaseSystemService):
    def register_policies(self) -> None:
        """
        Registers policies that checks if the system has a current outage

        """
        policy = HasSystemOutagePolicy(system_service=self)
        for command_class in self.app.domain.commands:
            self.app.register_policy(command_class, policy, EventTiming.BEFORE)

    def retrieve_outages(
        self, _cmd: command.RetrieveOutagesCommand
    ) -> list[model.Outage]:
        with self.repository.uow() as uow:
            outages = self.repository.crud(
                uow,
                None,
                model.Outage,
                None,
                None,
                CrudOperation.READ_ALL,
            )
        return outages



from datetime import timedelta
from typing import Any, Type

from cachetools import TTLCache, cached

from gen_epix.casedb.domain import command, exc, model
from gen_epix.casedb.domain.service.organization import BaseOrganizationService
from gen_epix.fastapp import Command, CrudOperation
from gen_epix.fastapp.model import CrudCommand


class OrganizationService(BaseOrganizationService):
    DEFAULT_CFG = {
        "user_invitation_time_to_live": 86400,  # 1 day
    }

    CACHE_INVALIDATION_COMMANDS: tuple[Type[Command], ...] = (
        command.UserCrudCommand,
        command.UpdateUserCommand,
    )

    def crud(
        self,
        cmd: CrudCommand,
    ) -> Any:
        retval = super().crud(cmd)
        # Invalidate cache
        if issubclass(type(cmd), OrganizationService.CACHE_INVALIDATION_COMMANDS):
            self.retrieve_user_by_key.cache_clear()  # type:ignore[attr-defined]
        return retval

    @cached(cache=TTLCache(maxsize=1000, ttl=60))
    def retrieve_user_by_key(self, user_key: str) -> model.User:
        with self.repository.uow() as uow:
            return self.repository.retrieve_user_by_key(uow, user_key)

    def invite_user(
        self,
        cmd: command.InviteUserCommand,
    ) -> model.UserInvitation:
        user = cmd.user
        if user is None:
            raise exc.UnauthorizedAuthError("Command has no user")
        if user.id is None:
            raise exc.UnauthorizedAuthError("User has no ID")
        email = cmd.email
        initial_roles = cmd.roles
        organization_id = cmd.organization_id

        with self.repository.uow() as uow:
            # Verify if user already exists
            is_existing_user = self.app.user_manager.is_existing_user_by_key(
                cmd.email, uow
            )

            if is_existing_user:
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "acba1a0e",
                            f"User {email} already exists",
                        )
                    )
                raise exc.UserAlreadyExistsAuthError("User already exists")

            is_existing_organization = self.repository.crud(
                uow,
                user.id,
                model.Organization,
                None,
                organization_id,
                CrudOperation.EXISTS_ONE,
            )
            if not is_existing_organization:
                if self._logger:
                    self._logger.info(
                        self.create_log_message(
                            "cdf1b633",
                            f"Organization id {organization_id} does not exist",
                        )
                    )
                raise exc.InvalidIdsError("Organization does not exist")

            # Verify if invitation(s) already exists for this email, and delete those
            # TODO: Must be done within the same session to be safe,
            # so requires specific repository method
            user_invitations: list[model.UserInvitation] = self.repository.crud(
                uow,
                user.id,
                model.UserInvitation,
                None,
                None,
                CrudOperation.READ_ALL,
            )  # type: ignore
            user_invitations = [x for x in user_invitations if x.email == email]
            if user_invitations:
                self.repository.crud(
                    uow,
                    user.id,
                    model.UserInvitation,
                    None,
                    [x.id for x in user_invitations],  # type: ignore
                    CrudOperation.DELETE_SOME,
                )  # type: ignore
            # Create user invitation
            user_invitation = model.UserInvitation(
                id=self.generate_id(),
                email=email,
                roles=initial_roles,
                organization_id=organization_id,
                invited_by_user_id=user.id,
                token=self.generate_user_invitation_token(),
                expires_at=self.generate_timestamp()
                + timedelta(
                    seconds=self.props.get(
                        "user_invitation_time_to_live",
                        OrganizationService.DEFAULT_CFG["user_invitation_time_to_live"],
                    )
                ),
            )
            user_invitation_in_db: model.UserInvitation = self.repository.crud(  # type: ignore[assignment]
                uow,
                user.id,
                model.UserInvitation,
                user_invitation,
                None,
                CrudOperation.CREATE_ONE,
            )
        return user_invitation_in_db

    def register_invited_user(
        self, cmd: command.RegisterInvitedUserCommand
    ) -> model.User:
        new_user = cmd.user
        if new_user is None:
            # Should not happen
            raise AssertionError("Command has no user")
        if not self.app.user_manager:
            raise exc.InvalidArgumentsError("User manager not set")

        with self.repository.uow() as uow:
            # Get possible user invitations
            user_invitations: list[model.UserInvitation] = self.repository.crud(  # type: ignore
                uow,
                None,
                model.UserInvitation,
                None,
                None,
                CrudOperation.READ_ALL,
            )
            now = self.generate_timestamp()
            user_invitations = [
                x
                for x in user_invitations
                if x.email == new_user.email and x.expires_at > now
            ]
            if not user_invitations:
                raise exc.UnauthorizedAuthError(
                    f"No valid invitations found for user {new_user.email}",
                )
            user_invitations_with_token = [
                x for x in user_invitations if x.token == cmd.token
            ]
            if not user_invitations_with_token:
                raise exc.UnauthorizedAuthError(
                    f"No invitation found for token {cmd.token}"
                )
            # Choose the invitation with the latest expiry date
            user_invitation: model.UserInvitation = sorted(
                user_invitations_with_token, key=lambda x: x.expires_at
            )[-1]
            # Set roles of the user
            new_user.roles = user_invitation.roles
            # Set ID and organization ID of the user
            new_user.organization_id = user_invitation.organization_id
            # Create user
            user_in_db = self.app.user_manager.create_new_user_from_token(
                new_user,
                user_invitation.token,
                created_by_user_id=user_invitation.invited_by_user_id,  # type: ignore[arg-type]
                roles=user_invitation.roles,  # type: ignore[arg-type]
            )
        return user_in_db  # type: ignore

    def update_user(
        self,
        cmd: command.UpdateUserCommand,
    ) -> model.User:
        if cmd.roles is not None and len(cmd.roles) == 0:
            raise exc.InvalidArgumentsError("Roles cannot be empty")
        with self.repository.uow() as uow:
            tgt_user = self.repository.crud(
                uow,
                cmd.user,
                model.User,
                None,
                cmd.tgt_user_id,
                CrudOperation.READ_ONE,
            )
            assert isinstance(tgt_user, model.User)
            is_active = tgt_user.is_active if cmd.is_active is None else cmd.is_active
            roles = tgt_user.roles if cmd.roles is None else cmd.roles
            organization_id = (
                tgt_user.organization_id
                if cmd.organization_id is None
                else cmd.organization_id
            )
            # Special case: no updates
            if (
                tgt_user.is_active == is_active
                and tgt_user.roles == roles
                and tgt_user.organization_id == organization_id
            ):
                return tgt_user
            # Check if organization_id exists
            if tgt_user.organization_id != organization_id:
                self.repository.crud(
                    uow,
                    cmd.user,
                    model.Organization,
                    None,
                    organization_id,
                    CrudOperation.READ_ONE,
                )
            # Update user
            tgt_user.is_active = is_active
            tgt_user.roles = roles
            tgt_user.organization_id = organization_id
            tgt_user = self.repository.crud(
                uow,
                cmd.user,
                model.User,
                tgt_user,
                None,
                CrudOperation.UPDATE_ONE,
            )
            assert isinstance(tgt_user, model.User)

        # Invalidate cache for the user
        self.retrieve_user_by_key.cache_clear()
        return tgt_user

    def retrieve_organization_contact(
        self,
        cmd: command.RetrieveOrganizationContactCommand,
    ) -> list[model.Contact]:
        if (
            sum(
                [
                    cmd.organization_ids is not None,
                    cmd.site_ids is not None,
                    cmd.contact_ids is not None,
                ]
            )
            != 1
        ):
            raise exc.InvalidArgumentsError(
                "Exactly one of organization_ids, site_ids or contact_ids must be provided"
            )
        user, repository = self._get_user_and_repository(cmd)

        organizations: list[model.Organization]
        sites: list[model.Site]
        contacts: list[model.Contact]
        with repository.uow() as uow:
            if cmd.organization_ids:
                organizations = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Organization,
                    None,
                    cmd.organization_ids,
                    CrudOperation.READ_SOME,
                    cascade_read=True,  # type: ignore[arg-type]
                )
                sites = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Site,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    cascade_read=True,  # type: ignore[arg-type]
                )
                contacts = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Contact,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    cascade_read=False,  # type: ignore[arg-type]
                )
                organization_ids = set(cmd.organization_ids)
                sites = [x for x in sites if x.organization_id in organization_ids]
                site_ids = {x.id for x in sites}
                contacts = [x for x in contacts if x.site_id in site_ids]
            elif cmd.site_ids:
                sites = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Site,
                    None,
                    cmd.site_ids,
                    CrudOperation.READ_SOME,
                    cascade_read=True,  # type: ignore[arg-type]
                )
                organizations = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Organization,
                    None,
                    list({x.organization_id for x in sites}),
                    CrudOperation.READ_SOME,
                    cascade_read=True,  # type: ignore[arg-type]
                )
                contacts = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Contact,
                    None,
                    None,
                    CrudOperation.READ_ALL,
                    cascade_read=False,  # type: ignore[arg-type]
                )
                site_ids = {x.id for x in sites}
                contacts = [x for x in contacts if x.site_id in site_ids]
            elif cmd.contact_ids:
                contacts = (
                    repository.crud(  # type: ignore[assignment]
                        uow,
                        user.id,
                        model.Contact,
                        None,
                        cmd.contact_ids,
                        CrudOperation.READ_ALL,
                        cascade_read=True,  # type: ignore[arg-type]
                        links=self.app.domain.get_model_links(
                            model.Contact, service_type=self.service_type
                        ),
                    ),
                )
                sites = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Site,
                    None,
                    list({x.site_id for x in contacts}),
                    CrudOperation.READ_SOME,
                    cascade_read=True,  # type: ignore[arg-type]
                )
                organizations = repository.crud(  # type: ignore[assignment]
                    uow,
                    user.id,
                    model.Organization,
                    None,
                    list({x.organization_id for x in sites}),
                    CrudOperation.READ_SOME,
                    cascade_read=True,  # type: ignore[arg-type]
                )
            else:
                raise AssertionError

        # Replace organization and site with cascaded version
        site_map = {x.id: x for x in sites}
        organization_map = {x.id: x for x in organizations}
        for site in site_map.values():
            site.organization = organization_map[site.organization_id]
        for contact in contacts:
            contact.site = site_map[contact.site_id]
        return contacts



from gen_epix.fastapp import PermissionTypeSet
from gen_epix.fastapp.services.rbac import BaseRbacService
from gen_epix.seqdb.domain import command
from gen_epix.seqdb.domain.enum import Role


class RoleGenerator:

    ROLE_PERMISSIONS = {
        Role.ADMIN: {  # type: ignore
            # organization
            (command.IdentifierIssuerCrudCommand, PermissionTypeSet.CU),
            (command.UserCrudCommand, PermissionTypeSet.R),
            (command.UserInvitationCrudCommand, PermissionTypeSet.CRD),
            (command.InviteUserCommand, PermissionTypeSet.E),
            (command.UpdateUserCommand, PermissionTypeSet.E),
            (command.OutageCrudCommand, PermissionTypeSet.CRUD),
            (command.DataCollectionCrudCommand, PermissionTypeSet.CU),
            (command.DataCollectionSetCrudCommand, PermissionTypeSet.CRUD),
            (command.DataCollectionSetMemberCrudCommand, PermissionTypeSet.CRUD),
        },
        Role.REFDATA_ADMIN: {  # type: ignore
            # organization
            # seq.metadata CRUD commands
            (command.AlignmentProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.AssemblyProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.AstProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpCrudCommand, PermissionTypeSet.CRU),
            (command.SnpDetectionProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpSetCrudCommand, PermissionTypeSet.CRU),
            (command.RefSnpSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.LibraryPrepProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.LocusCrudCommand, PermissionTypeSet.CRU),
            (command.LocusDetectionProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.LocusSetCrudCommand, PermissionTypeSet.CRU),
            (command.LocusSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.RefAlleleCrudCommand, PermissionTypeSet.CRU),
            (command.PcrProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.RefSeqCrudCommand, PermissionTypeSet.CRU),
            (command.SeqCategoryCrudCommand, PermissionTypeSet.CRU),
            (command.SeqCategorySetCrudCommand, PermissionTypeSet.CRU),
            (command.SeqClassificationProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.SeqDistanceProtocolCrudCommand, PermissionTypeSet.CRU),
            (command.SubtypingSchemeCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonLocusLinkCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonSetCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonSetMemberCrudCommand, PermissionTypeSet.CRU),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.CRU),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.CRU),
            (command.TaxonomyProtocolCrudCommand, PermissionTypeSet.CRU),
        },
        Role.DATA_WRITER: {  # type: ignore
            # seq.persistable CRUD commands
            (command.AlleleCrudCommand, PermissionTypeSet.CUD),
            (command.AlleleAlignmentCrudCommand, PermissionTypeSet.CUD),
            (command.SampleCrudCommand, PermissionTypeSet.CUD),
            (command.ReadSetCrudCommand, PermissionTypeSet.CUD),
            (command.SeqCrudCommand, PermissionTypeSet.CUD),
            (command.AlleleProfileCrudCommand, PermissionTypeSet.CUD),
            (command.SeqAlignmentCrudCommand, PermissionTypeSet.CUD),
            (command.SeqTaxonomyCrudCommand, PermissionTypeSet.CUD),
            (command.PcrMeasurementCrudCommand, PermissionTypeSet.CUD),
            (command.AstMeasurementCrudCommand, PermissionTypeSet.CUD),
            (command.AstPredictionCrudCommand, PermissionTypeSet.CUD),
            (command.SeqDistanceCrudCommand, PermissionTypeSet.CUD),
            (command.SeqClassificationCrudCommand, PermissionTypeSet.CUD),
            (command.SnpProfileCrudCommand, PermissionTypeSet.CUD),
        },
        Role.DATA_READER: {  # type: ignore
            # organization
            (command.DataCollectionCrudCommand, PermissionTypeSet.R),
            # seq.metadata CRUD commands
            (command.AlignmentProtocolCrudCommand, PermissionTypeSet.R),
            (command.AssemblyProtocolCrudCommand, PermissionTypeSet.R),
            (command.AstProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSnpCrudCommand, PermissionTypeSet.R),
            (command.SnpDetectionProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSnpSetCrudCommand, PermissionTypeSet.R),
            (command.RefSnpSetMemberCrudCommand, PermissionTypeSet.R),
            (command.LibraryPrepProtocolCrudCommand, PermissionTypeSet.R),
            (command.LocusCrudCommand, PermissionTypeSet.R),
            (command.LocusDetectionProtocolCrudCommand, PermissionTypeSet.R),
            (command.LocusSetCrudCommand, PermissionTypeSet.R),
            (command.LocusSetMemberCrudCommand, PermissionTypeSet.R),
            (command.PcrProtocolCrudCommand, PermissionTypeSet.R),
            (command.RefSeqCrudCommand, PermissionTypeSet.R),
            (command.SeqCategoryCrudCommand, PermissionTypeSet.R),
            (command.SeqCategorySetCrudCommand, PermissionTypeSet.R),
            (command.SeqClassificationProtocolCrudCommand, PermissionTypeSet.R),
            (command.SeqDistanceProtocolCrudCommand, PermissionTypeSet.R),
            (command.SubtypingSchemeCrudCommand, PermissionTypeSet.R),
            (command.TaxonCrudCommand, PermissionTypeSet.R),
            (command.TaxonLocusLinkCrudCommand, PermissionTypeSet.R),
            (command.TaxonSetCrudCommand, PermissionTypeSet.R),
            (command.TaxonSetMemberCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmClassCrudCommand, PermissionTypeSet.R),
            (command.TreeAlgorithmCrudCommand, PermissionTypeSet.R),
            (command.TaxonomyProtocolCrudCommand, PermissionTypeSet.R),
            # seq CRUD commands
            (command.AlleleCrudCommand, PermissionTypeSet.R),
            (command.AlleleAlignmentCrudCommand, PermissionTypeSet.R),
            (command.SampleCrudCommand, PermissionTypeSet.R),
            (command.ReadSetCrudCommand, PermissionTypeSet.R),
            (command.SeqCrudCommand, PermissionTypeSet.R),
            (command.AlleleProfileCrudCommand, PermissionTypeSet.R),
            (command.SeqAlignmentCrudCommand, PermissionTypeSet.R),
            (command.SeqTaxonomyCrudCommand, PermissionTypeSet.R),
            (command.PcrMeasurementCrudCommand, PermissionTypeSet.R),
            (command.AstMeasurementCrudCommand, PermissionTypeSet.R),
            (command.AstPredictionCrudCommand, PermissionTypeSet.R),
            (command.SeqDistanceCrudCommand, PermissionTypeSet.R),
            (command.SeqClassificationCrudCommand, PermissionTypeSet.R),
            (command.SnpProfileCrudCommand, PermissionTypeSet.R),
            # seq non-CRUD commands
            (command.RetrievePhylogeneticTreeCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteAlleleProfileCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSnpProfileCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteContigCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSampleCommand, PermissionTypeSet.E),
            (command.RetrieveCompleteSeqCommand, PermissionTypeSet.E),
        },
        Role.GUEST: set(),  # type: ignore
    }

    # Tree hierarchy of roles: each role can do everything the roles below it can do.
    # Hierarchy described here per role with union of all roles below it.
    ROLE_HIERARCHY = {
        Role.ROOT: {  # type: ignore
            Role.ADMIN,
            Role.REFDATA_ADMIN,
            Role.DATA_READER,
            Role.DATA_WRITER,
            Role.GUEST,
        },
        Role.ADMIN: {  # type: ignore
            Role.REFDATA_ADMIN,
            Role.DATA_READER,
            Role.DATA_WRITER,
            Role.GUEST,
        },
        Role.REFDATA_ADMIN: {Role.GUEST},  # type: ignore
        Role.DATA_READER: {Role.GUEST},  # type: ignore
        Role.DATA_WRITER: {Role.DATA_READER, Role.GUEST},  # type: ignore
        Role.GUEST: set(),  # type: ignore
    }

    ROLE_PERMISSIONS = BaseRbacService.expand_hierarchical_role_permissions(
        ROLE_HIERARCHY, ROLE_PERMISSIONS
    )



# pylint: disable=wildcard-import, unused-wildcard-import

from __future__ import annotations

from gen_epix.seqdb.domain import DOMAIN, enum
from gen_epix.seqdb.repositories.sa_model.base import RowMetadataMixin
from gen_epix.seqdb.repositories.sa_model.organization import *
from util.util import set_entity_repository_model_classes

set_entity_repository_model_classes(
    DOMAIN, enum.ServiceType, RowMetadataMixin, "gen_epix.seqdb.repositories.sa_model"
)



from typing import Any

from gen_epix.casedb.domain import command, enum, exc
from gen_epix.casedb.domain.policy.abac import BaseReadSelfResultsOnlyPolicy
from gen_epix.fastapp import Command, CrudOperation, CrudOperationSet


class ReadSelfResultsOnlyPolicy(BaseReadSelfResultsOnlyPolicy):
    def filter(self, cmd: Command, retval: Any) -> Any:
        if not cmd.user or not cmd.user.id:
            raise exc.ServiceException("Command has no user")
        # TODO: replace filter for AFTER with injecting a filter DURING for efficiency
        if not isinstance(cmd, command.CrudCommand):
            raise NotImplementedError
        if cmd.operation not in CrudOperationSet.READ_OR_EXISTS.value:
            # Policy only applies to read or exists operations
            return retval

        # No restrictions for ORG_ADMIN, APP_ADMIN users and above
        is_not_restricted = (
            len(cmd.user.roles.intersection(enum.RoleSet.GE_ORG_ADMIN.value)) > 0
        )
        if is_not_restricted:
            return retval

        # Filter results based on organizations
        is_read_all = cmd.operation == CrudOperation.READ_ALL
        is_read_one = cmd.operation == CrudOperation.READ_ONE
        msg = "No data for user"
        user_id = cmd.user.id
        if isinstance(cmd, command.UserCrudCommand):
            if is_read_all:
                retval = [x for x in retval if x.id == user_id]
            if is_read_one and retval.id != user_id:
                raise exc.UnauthorizedAuthError(msg)
            if not is_read_one and any(x.id != user_id for x in retval):
                raise exc.UnauthorizedAuthError(msg)
        elif isinstance(cmd, command.UserAccessCasePolicyCrudCommand) or isinstance(
            cmd, command.UserShareCasePolicyCrudCommand
        ):
            if is_read_all:
                retval = [x for x in retval if x.user_id == user_id]
            if is_read_one and retval.user_id != user_id:
                raise exc.UnauthorizedAuthError(msg)
            if not is_read_one and any(x.user_id != user_id for x in retval):
                raise exc.UnauthorizedAuthError(msg)
        else:
            raise NotImplementedError
        return retval



import abc

from gen_epix.casedb.domain import model  # forces models to be registered now
from gen_epix.casedb.domain import DOMAIN
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.fastapp import BaseRepository, BaseUnitOfWork


class BaseOrganizationRepository(BaseRepository):
    ENTITIES = DOMAIN.get_dag_sorted_entities(
        service_type=ServiceType.ORGANIZATION, persistable=True
    )

    @abc.abstractmethod
    def is_existing_user_by_key(
        self, uow: BaseUnitOfWork, user_key: str | None
    ) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_user_by_key(self, uow: BaseUnitOfWork, user_key: str) -> model.User:
        raise NotImplementedError()



import abc
from typing import Type
from uuid import UUID

from gen_epix.casedb.domain import command, model
from gen_epix.casedb.domain.enum import ServiceType
from gen_epix.casedb.domain.repository.case import BaseCaseRepository
from gen_epix.fastapp import BaseService


class BaseCaseService(BaseService):
    SERVICE_TYPE = ServiceType.CASE

    NO_ABAC_COMMAND_CLASSES: set[Type[command.Command]] = {
        command.TreeAlgorithmClassCrudCommand,
        command.TreeAlgorithmCrudCommand,
        command.GeneticDistanceProtocolCrudCommand,
        command.DimCrudCommand,
        command.ColCrudCommand,
        command.CaseTypeSetCategoryCrudCommand,
        command.CaseSetCategoryCrudCommand,
        command.CaseSetStatusCrudCommand,
    }
    ABAC_METADATA_COMMAND_CLASSES: set[Type[command.Command]] = {
        command.CaseTypeCrudCommand,
        command.CaseTypeSetMemberCrudCommand,
        command.CaseTypeSetCrudCommand,
        command.CaseTypeColCrudCommand,
        command.CaseTypeColSetMemberCrudCommand,
        command.CaseTypeColSetCrudCommand,
    }
    ABAC_DATA_COMMAND_CLASSES: set[Type[command.Command]] = {
        command.CaseCrudCommand,
        command.CaseSetCrudCommand,
        command.CaseSetMemberCrudCommand,
        command.CaseDataCollectionLinkCrudCommand,
        command.CaseSetDataCollectionLinkCrudCommand,
    }
    CASCADE_DELETE_MODEL_CLASSES: dict[
        Type[model.Model], tuple[Type[model.Model], ...]
    ] = {
        model.CaseTypeSet: (model.CaseTypeSetMember,),
        model.CaseType: (model.CaseTypeSetMember,),
        model.CaseTypeColSet: (model.CaseTypeColSetMember,),
        model.CaseTypeCol: (model.CaseTypeColSetMember,),
        model.CaseSet: (
            model.CaseSetDataCollectionLink,
            model.CaseSetMember,
        ),
        model.Case: (
            model.CaseDataCollectionLink,
            model.CaseSetMember,
        ),
    }

    # Property overridden to provide narrower return value to support linter
    @property  # type: ignore
    def repository(self) -> BaseCaseRepository:  # type: ignore
        return super().repository  # type: ignore

    def register_handlers(self) -> None:
        f = self.app.register_handler
        for command_class in self.app.domain.get_crud_commands_for_service_type(
            self.service_type
        ):
            f(command_class, self.crud)
        f(command.CaseSetCreateCommand, self.create_cases_or_set)
        f(command.CasesCreateCommand, self.create_cases_or_set)
        f(command.RetrieveCompleteCaseTypeCommand, self.retrieve_complete_case_type)
        f(command.RetrieveCaseTypeStatsCommand, self.retrieve_case_type_stats)
        f(command.RetrieveCaseSetStatsCommand, self.retrieve_case_set_stats)
        f(command.RetrieveCasesByQueryCommand, self.retrieve_cases_by_query)
        f(command.RetrieveCasesByIdCommand, self.retrieve_cases_by_id)
        f(command.RetrieveCaseRightsCommand, self.retrieve_case_or_set_rights)
        f(command.RetrieveCaseSetRightsCommand, self.retrieve_case_or_set_rights)
        f(command.RetrieveGeneticSequenceByCaseCommand, self.retrieve_genetic_sequence)
        f(
            command.RetrievePhylogeneticTreeByCasesCommand,
            self.retrieve_phylogenetic_tree,
        )
        f(command.RetrieveGeneticSequenceByCaseCommand, self.retrieve_genetic_sequence)

    @abc.abstractmethod
    def create_cases_or_set(
        self, cmd: command.CaseSetCreateCommand | command.CasesCreateCommand
    ) -> model.CaseSet | list[model.Case] | None:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_complete_case_type(
        self,
        cmd: command.RetrieveCompleteCaseTypeCommand,
    ) -> model.CompleteCaseType:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_set_stats(
        self,
        cmd: command.RetrieveCaseSetStatsCommand,
    ) -> list[model.CaseSetStat]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_type_stats(
        self,
        cmd: command.RetrieveCaseTypeStatsCommand,
    ) -> list[model.CaseTypeStat]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_cases_by_query(
        self, cmd: command.RetrieveCasesByQueryCommand
    ) -> list[UUID]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_cases_by_id(
        self, cmd: command.RetrieveCasesByIdCommand
    ) -> list[model.Case]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_case_or_set_rights(
        self,
        cmd: command.RetrieveCaseRightsCommand | command.RetrieveCaseSetRightsCommand,
    ) -> list[model.CaseRights] | list[model.CaseSetRights]:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_phylogenetic_tree(
        self,
        cmd: command.RetrievePhylogeneticTreeByCasesCommand,
    ) -> model.PhylogeneticTree:
        raise NotImplementedError()

    @abc.abstractmethod
    def retrieve_genetic_sequence(
        self,
        cmd: command.RetrieveGeneticSequenceByCaseCommand,
    ) -> list[model.GeneticSequence]:
        raise NotImplementedError()



from typing import Any

from gen_epix.casedb.domain import command, enum, exc, model
from gen_epix.casedb.domain.policy.abac import BaseReadOrganizationResultsOnlyPolicy
from gen_epix.fastapp import Command, CrudOperation, CrudOperationSet


class ReadOrganizationResultsOnlyPolicy(BaseReadOrganizationResultsOnlyPolicy):
    def filter(self, cmd: Command, retval: Any) -> Any:
        if not cmd.user or not cmd.user.id:
            raise exc.ServiceException("Command has no user")
        # TODO: replace filter for AFTER with injecting a filter DURING for efficiency
        if not isinstance(cmd, command.CrudCommand):
            raise NotImplementedError
        if cmd.operation not in CrudOperationSet.READ_OR_EXISTS.value:
            # Policy only applies to read or exists operations
            return retval
        # No restrictions for APP_ADMIN users and above
        is_not_restricted = (
            len(cmd.user.roles.intersection(enum.RoleSet.GE_APP_ADMIN.value)) > 0
        )
        if is_not_restricted:
            return retval
        # Get organizations to filter on: user's own organization plus any
        # organizations they are admin for
        organization_ids = self.abac_service.get_organizations_under_admin(cmd.user)
        if organization_ids:
            organization_ids.add(cmd.user.organization_id)
        else:
            organization_ids = {cmd.user.organization_id}
        # Filter results based on organizations
        is_read_all = cmd.operation == CrudOperation.READ_ALL
        is_read_one = cmd.operation == CrudOperation.READ_ONE
        msg1 = "User is not an admin for the organization and/or does not belong to it"
        msg2 = "User is not an admin for some of the organizations and/or does not belong to them"
        if (
            isinstance(cmd, command.UserCrudCommand)
            or isinstance(cmd, command.OrganizationAdminPolicyCrudCommand)
            or isinstance(cmd, command.OrganizationAccessCasePolicyCrudCommand)
            or isinstance(cmd, command.OrganizationShareCasePolicyCrudCommand)
            or isinstance(cmd, command.UserInvitationCrudCommand)
        ):
            if is_read_all:
                retval = [x for x in retval if x.organization_id in organization_ids]
            if is_read_one and retval.organization_id not in organization_ids:
                raise exc.UnauthorizedAuthError(msg1)
            if not is_read_one and any(
                x.organization_id not in organization_ids for x in retval
            ):
                raise exc.UnauthorizedAuthError(msg2)
        elif isinstance(cmd, command.UserAccessCasePolicyCrudCommand) or isinstance(
            cmd, command.UserShareCasePolicyCrudCommand
        ):
            if is_read_one:
                user_ids = [cmd.objs]
            elif is_read_all:
                user_ids = None
            else:
                user_ids = list(set(cmd.objs))
            users: list[model.User] = self.abac_service.app.handle(
                command.UserCrudCommand(
                    user=cmd.user,
                    objs=None,
                    obj_ids=user_ids,
                    operation=(
                        CrudOperation.READ_ALL
                        if is_read_all
                        else CrudOperation.READ_SOME
                    ),
                )
            )
            valid_user_ids = {
                x.id for x in users if x.organization_id in organization_ids
            }
            if is_read_all:
                retval = [x for x in retval if x.user_id in valid_user_ids]
            else:
                if is_read_one and retval.user_id not in valid_user_ids:
                    raise exc.UnauthorizedAuthError(msg1)
                if not is_read_one and not {x.user_id for x in retval}.issubset(
                    valid_user_ids
                ):
                    raise exc.UnauthorizedAuthError(msg2)
        else:
            raise NotImplementedError
        return retval



from typing import ClassVar
from uuid import UUID

from pydantic import Field

from gen_epix.casedb.domain import DOMAIN, enum
from gen_epix.casedb.domain.model.abac.case_abac import (
    CaseTypeAccessAbac,
    CaseTypeShareAbac,
)
from gen_epix.casedb.domain.model.case.case import (
    _ENTITY_KWARGS,
    CaseType,
    CaseTypeCol,
    CaseTypeDim,
    Col,
    Dim,
    GeneticDistanceProtocol,
    TreeAlgorithm,
)
from gen_epix.casedb.domain.model.ontology import EtiologicalAgent, Etiology
from gen_epix.fastapp.domain import Entity

_SERVICE_TYPE = enum.ServiceType.CASE
_ENTITY_KWARGS = {
    "schema_name": _SERVICE_TYPE.value.lower(),
}


class CompleteCaseType(CaseType):
    ENTITY: ClassVar = Entity(
        snake_case_plural_name="complete_case_types",
        persistable=False,
        **_ENTITY_KWARGS,
    )
    etiologies: dict[UUID, Etiology] = Field(
        description="The etiologies used by the case type"
    )
    etiological_agents: dict[UUID, EtiologicalAgent] = Field(
        description="The etiological agents used by the case type"
    )
    dims: dict[UUID, Dim] = Field(description="The dimensions used by the case type")
    cols: dict[UUID, Col] = Field(description="The columns used by the case type")
    case_type_dims: list[CaseTypeDim] = Field(
        description="The ordered list of case type dimensions"
    )
    case_type_cols: dict[UUID, CaseTypeCol] = Field(  # type: ignore
        description="The case type columns for the case type"
    )
    case_type_col_order: list[UUID] = Field(
        description="The order of the case type columns outside the context of a dimension"
    )
    genetic_distance_protocols: dict[UUID, GeneticDistanceProtocol] = Field(
        description="The genetic distance protocols used by the case type"
    )
    tree_algorithms: dict[enum.TreeAlgorithmType, TreeAlgorithm] = Field(
        description="The tree algorithms used by the case type"
    )
    case_type_access_abacs: dict[UUID, CaseTypeAccessAbac] = Field(
        description="The case type access ABAC object by data collection ID"
    )
    case_type_share_abacs: dict[UUID, CaseTypeShareAbac] = Field(
        description="The case type share ABAC object by data collection ID"
    )


DOMAIN.register_locals(locals(), service_type=_SERVICE_TYPE)



# pylint: disable=too-few-public-methods
# This module defines base classes, methods are added later


from typing import ClassVar, Self
from uuid import UUID

from pydantic import model_validator

from gen_epix.fastapp.services import auth
from gen_epix.seqdb.domain import DOMAIN, enum
from gen_epix.seqdb.domain.command.base import Command


# auth
class GetIdentityProvidersCommand(auth.GetIdentityProvidersCommand):
    SERVICE_TYPE: ClassVar = enum.ServiceType.AUTH


# organization
class RetrieveCompleteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION


class InviteUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    email: str
    roles: set[enum.Role]


class RegisterInvitedUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION
    token: str


class UpdateUserCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.ORGANIZATION

    tgt_user_id: UUID
    is_active: bool | None
    roles: set[enum.Role] | None
    data_collection_ids: set[UUID] | None
    organization_id: UUID | None


# system
class RetrieveOutagesCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SYSTEM
    pass


class RetrieveCompleteContigCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ


class RetrieveCompleteAlleleProfileCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ


class RetrieveCompleteSnpProfileCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ


class RetrieveCompleteSeqCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ

    seq_ids: list[UUID]


class RetrieveCompleteSampleCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ


class RetrievePhylogeneticTreeCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ

    seq_distance_protocol_id: UUID
    tree_algorithm: enum.TreeAlgorithm
    seq_ids: list[UUID]
    leaf_names: list[str] | None

    @model_validator(mode="after")
    def _validate_state(self) -> Self:
        if self.leaf_names is not None and len(self.leaf_names) != len(self.seq_ids):
            raise ValueError(
                "leaf_codes must be None or have the same length as seq_ids"
            )
        return self


class RetrieveMultipleAlignmentCommand(Command):
    SERVICE_TYPE: ClassVar = enum.ServiceType.SEQ


DOMAIN.register_locals(locals())



from gen_epix.casedb.domain import command, enum, model
from gen_epix.casedb.domain.policy.abac import BaseIsOrganizationAdminPolicy
from gen_epix.fastapp import Command, CrudCommand, CrudOperation, CrudOperationSet


class IsOrganizationAdminPolicy(BaseIsOrganizationAdminPolicy):
    def is_allowed(self, cmd: Command) -> bool:
        if cmd.user is None:
            return False

        if cmd.user.roles.intersection(enum.RoleSet.GE_APP_ADMIN.value):
            return True

        if isinstance(cmd, CrudCommand):
            if cmd.operation not in CrudOperationSet.WRITE.value:
                # Policy only applies to write operations
                return True
            objs = cmd.objs
            if objs is None:
                objs = []
            elif not isinstance(objs, list):
                objs = [objs]
            # Determine the set of organizations affected
            if isinstance(
                cmd, command.OrganizationAccessCasePolicyCrudCommand
            ) or isinstance(cmd, command.OrganizationShareCasePolicyCrudCommand):
                organization_ids = set(x.organization_id for x in objs)  # type: ignore
            elif isinstance(cmd, command.UserAccessCasePolicyCrudCommand) or isinstance(
                cmd, command.UserShareCasePolicyCrudCommand
            ):
                with self.abac_service.repository.uow() as uow:
                    users: list[model.User] = self.abac_service.app.handle(
                        command.UserCrudCommand(
                            user=cmd.user,
                            objs=None,
                            obj_ids=list(set(x.user_id for x in objs)),
                            operation=CrudOperation.READ_SOME,
                        )
                    )
                    organization_ids = set(x.organization_id for x in users)
            elif isinstance(cmd, command.SiteCrudCommand):
                organization_ids = {x.organization_id for x in objs}
            elif isinstance(cmd, command.ContactCrudCommand):
                contacts = self.abac_service.app.handle(
                    command.ContactCrudCommand(
                        user=cmd.user,
                        objs=None,
                        obj_ids=list(set(x.contact_id for x in objs)),
                        operation=CrudOperation.READ_SOME,
                    )
                )
                sites = self.abac_service.app.handle(
                    command.ContactCrudCommand(
                        user=cmd.user,
                        objs=None,
                        obj_ids=list(set(x.site_id for x in contacts)),
                        operation=CrudOperation.READ_SOME,
                    )
                )
                organization_ids = {x.organization_id for x in sites}
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError
        # Check if user is an admin for all of the affected organizations
        user_admin_organization_ids = self.abac_service.get_organizations_under_admin(
            cmd.user
        )
        has_permission = organization_ids.issubset(user_admin_organization_ids)
        return has_permission
